# encoding: utf-8

"""Unit-test suite for pptx.table module"""

from __future__ import absolute_import, division, print_function, unicode_literals

import pytest

from pptx.dml.fill import FillFormat
from pptx.enum.text import MSO_ANCHOR
from pptx.oxml.ns import qn
from pptx.oxml.table import CT_Table, CT_TableCell, TcRange
from pptx.shapes.graphfrm import GraphicFrame
from pptx.table import (
    _Cell,
    _CellCollection,
    _Column,
    _ColumnCollection,
    _Row,
    _RowCollection,
    Table,
)
from pptx.text.text import TextFrame
from pptx.util import Inches, Length, Pt

from .unitutil.cxml import element, xml
from .unitutil.mock import call, class_mock, instance_mock, property_mock


class DescribeTable(object):
    def it_provides_access_to_its_cells(self, tbl_, tc_, _Cell_, cell_):
        row_idx, col_idx = 4, 2
        tbl_.tc.return_value = tc_
        _Cell_.return_value = cell_
        table = Table(tbl_, None)

        cell = table.cell(row_idx, col_idx)

        tbl_.tc.assert_called_once_with(row_idx, col_idx)
        _Cell_.assert_called_once_with(tc_, table)
        assert cell is cell_

    def it_provides_access_to_its_columns(self, columns_fixture):
        table, expected_columns_ = columns_fixture
        assert table.columns is expected_columns_

    def it_can_iterate_its_grid_cells(self, request, _Cell_):
        tbl = element("a:tbl/(a:tr/(a:tc,a:tc),a:tr/(a:tc,a:tc))")
        expected_tcs = tbl.xpath(".//a:tc")
        expected_cells = _Cell_.side_effect = [
            instance_mock(request, _Cell, name="cell%d" % idx) for idx in range(4)
        ]
        table = Table(tbl, None)

        cells = list(table.iter_cells())

        assert cells == expected_cells
        assert _Cell_.call_args_list == [call(tc, table) for tc in expected_tcs]

    def it_provides_access_to_its_rows(self, rows_fixture):
        table, expected_rows_ = rows_fixture
        assert table.rows is expected_rows_

    def it_updates_graphic_frame_width_on_width_change(self, dx_fixture):
        table, expected_width = dx_fixture
        table.notify_width_changed()
        assert table._graphic_frame.width == expected_width

    def it_updates_graphic_frame_height_on_height_change(self, dy_fixture):
        table, expected_height = dy_fixture
        table.notify_height_changed()
        assert table._graphic_frame.height == expected_height

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def columns_fixture(self, table, columns_):
        table._columns = columns_
        return table, columns_

    @pytest.fixture
    def dx_fixture(self, graphic_frame_):
        tbl_cxml = "a:tbl/a:tblGrid/(a:gridCol{w=111},a:gridCol{w=222})"
        table = Table(element(tbl_cxml), graphic_frame_)
        expected_width = 333
        return table, expected_width

    @pytest.fixture
    def dy_fixture(self, graphic_frame_):
        tbl_cxml = "a:tbl/(a:tr{h=100},a:tr{h=200})"
        table = Table(element(tbl_cxml), graphic_frame_)
        expected_height = 300
        return table, expected_height

    @pytest.fixture
    def rows_fixture(self, table, rows_):
        table._rows = rows_
        return table, rows_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _Cell_(self, request):
        return class_mock(request, "pptx.table._Cell")

    @pytest.fixture
    def cell_(self, request):
        return instance_mock(request, _Cell)

    @pytest.fixture
    def columns_(self, request):
        return instance_mock(request, _ColumnCollection)

    @pytest.fixture
    def graphic_frame_(self, request):
        return instance_mock(request, GraphicFrame)

    @pytest.fixture
    def rows_(self, request):
        return instance_mock(request, _RowCollection)

    @pytest.fixture
    def table(self):
        return Table(element("a:tbl"), None)

    @pytest.fixture
    def tbl_(self, request):
        return instance_mock(request, CT_Table)

    @pytest.fixture
    def tc_(self, request):
        return instance_mock(request, CT_TableCell)


class DescribeTableBooleanProperties(object):
    def it_knows_its_boolean_property_settings(self, boolprop_get_fixture):
        table, boolprop_name, expected_value = boolprop_get_fixture
        boolprop_value = getattr(table, boolprop_name)
        assert boolprop_value is expected_value

    def it_can_change_its_boolean_property_settings(self, boolprop_set_fixture):
        table, boolprop_name, new_value, expected_xml = boolprop_set_fixture
        setattr(table, boolprop_name, new_value)
        assert table._tbl.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("a:tbl", "first_row", False),
            ("a:tbl/a:tblPr", "first_row", False),
            ("a:tbl/a:tblPr{firstRow=1}", "first_row", True),
            ("a:tbl/a:tblPr{firstRow=0}", "first_row", False),
            ("a:tbl/a:tblPr{firstRow=true}", "first_row", True),
            ("a:tbl/a:tblPr{firstRow=false}", "first_row", False),
            ("a:tbl/a:tblPr{firstCol=1}", "first_col", True),
            ("a:tbl/a:tblPr{lastRow=0}", "last_row", False),
            ("a:tbl/a:tblPr{lastCol=true}", "last_col", True),
            ("a:tbl/a:tblPr{bandRow=false}", "horz_banding", False),
            ("a:tbl/a:tblPr", "vert_banding", False),
        ]
    )
    def boolprop_get_fixture(self, request):
        tbl_cxml, boolprop_name, expected_value = request.param
        table = Table(element(tbl_cxml), None)
        return table, boolprop_name, expected_value

    @pytest.fixture(
        params=[
            ("a:tbl", "first_row", True, "a:tbl/a:tblPr{firstRow=1}"),
            ("a:tbl", "first_row", False, "a:tbl/a:tblPr"),
            ("a:tbl/a:tblPr", "first_row", True, "a:tbl/a:tblPr{firstRow=1}"),
            ("a:tbl/a:tblPr", "first_row", False, "a:tbl/a:tblPr"),
            (
                "a:tbl/a:tblPr{firstRow=true}",
                "first_row",
                True,
                "a:tbl/a:tblPr{firstRow=1}",
            ),
            ("a:tbl/a:tblPr{firstRow=false}", "first_row", False, "a:tbl/a:tblPr"),
            (
                "a:tbl/a:tblPr{bandRow=1}",
                "first_row",
                True,
                "a:tbl/a:tblPr{bandRow=1,firstRow=1}",
            ),
            ("a:tbl", "first_col", True, "a:tbl/a:tblPr{firstCol=1}"),
            ("a:tbl", "last_row", True, "a:tbl/a:tblPr{lastRow=1}"),
            ("a:tbl", "last_col", True, "a:tbl/a:tblPr{lastCol=1}"),
            ("a:tbl", "horz_banding", True, "a:tbl/a:tblPr{bandRow=1}"),
            ("a:tbl", "vert_banding", True, "a:tbl/a:tblPr{bandCol=1}"),
        ]
    )
    def boolprop_set_fixture(self, request):
        tbl_cxml, boolprop_name, new_value, expected_tbl_cxml = request.param
        table = Table(element(tbl_cxml), None)
        expected_xml = xml(expected_tbl_cxml)
        return table, boolprop_name, new_value, expected_xml


class Describe_Cell(object):
    """Unit-test suite for `pptx.table._Cell` object."""

    def it_is_equal_to_other_instance_having_same_tc(self):
        tc = element("a:tc")
        other_tc = element("a:tc")
        cell = _Cell(tc, None)
        cell_with_same_tc = _Cell(tc, None)
        cell_with_other_tc = _Cell(other_tc, None)

        assert cell == cell_with_same_tc
        assert cell != cell_with_other_tc

    def it_has_a_fill(self, fill_fixture):
        cell = fill_fixture
        assert isinstance(cell.fill, FillFormat)

    def it_knows_whether_it_is_merge_origin_cell(self, origin_fixture):
        tc, expected_value = origin_fixture
        cell = _Cell(tc, None)

        is_merge_origin = cell.is_merge_origin

        assert is_merge_origin is expected_value

    def it_knows_whether_it_is_spanned(self, spanned_fixture):
        tc, expected_value = spanned_fixture
        cell = _Cell(tc, None)

        is_spanned = cell.is_spanned

        assert is_spanned is expected_value

    def it_knows_its_margin_settings(self, margin_get_fixture):
        cell, margin_prop_name, expected_value = margin_get_fixture
        margin_value = getattr(cell, margin_prop_name)
        assert margin_value == expected_value

    def it_can_change_its_margin_settings(self, margin_set_fixture):
        cell, margin_prop_name, new_value, expected_xml = margin_set_fixture
        setattr(cell, margin_prop_name, new_value)
        assert cell._tc.xml == expected_xml

    def it_raises_on_margin_assigned_other_than_int_or_None(
        self, margin_raises_fixture
    ):
        cell, margin_attr_name, val_of_invalid_type = margin_raises_fixture
        with pytest.raises(TypeError):
            setattr(cell, margin_attr_name, val_of_invalid_type)

    def it_can_merge_a_range_of_cells(self, TcRange_, tc_range_):
        tbl = element("a:tbl/(a:tr/(a:tc,a:tc),a:tr/(a:tc,a:tc))")
        tc, other_tc = tbl.tc(0, 0), tbl.tc(1, 1)
        TcRange_.return_value = tc_range_
        tc_range_.contains_merged_cell = False
        tc_range_.dimensions = 2, 2

        def tcs(*rowcols):
            return (tbl.tc(*rowcol) for rowcol in rowcols)

        tc_range_.iter_top_row_tcs.return_value = tcs((0, 0), (0, 1))
        tc_range_.iter_left_col_tcs.return_value = tcs((0, 0), (1, 0))
        tc_range_.iter_except_left_col_tcs.return_value = tcs((0, 1), (1, 1))
        tc_range_.iter_except_top_row_tcs.return_value = tcs((1, 0), (1, 1))
        expected_xml = xml(
            "a:tbl/(a:tr/(a:tc{gridSpan=2,rowSpan=2},a:tc{rowSpan=2,hMerge=1"
            "}),a:tr/(a:tc{gridSpan=2,vMerge=1},a:tc{hMerge=1,vMerge=1}))"
        )
        cell, other_cell = _Cell(tc, None), _Cell(other_tc, None)

        cell.merge(other_cell)

        TcRange_.assert_called_once_with(tc, other_tc)
        tc_range_.move_content_to_origin.assert_called_once_with()
        assert tbl.xml == expected_xml

    def but_it_raises_when_cells_are_from_different_tables(self, TcRange_, tc_range_):
        TcRange_.return_value = tc_range_
        tc_range_.in_same_table = False
        cell, other_cell = _Cell(None, None), _Cell(None, None)

        with pytest.raises(ValueError) as e:
            cell.merge(other_cell)
        assert "different table" in str(e.value)

    def and_it_raises_when_range_contains_merged_cell(self, TcRange_, tc_range_):
        TcRange_.return_value = tc_range_
        tc_range_.contains_merged_cell = True
        cell, other_cell = _Cell(None, None), _Cell(None, None)

        with pytest.raises(ValueError) as e:
            cell.merge(other_cell)
        assert "contains one or more merged cells" in str(e.value)

    def it_knows_how_many_rows_the_merge_spans(self, height_fixture):
        tc, expected_value = height_fixture
        cell = _Cell(tc, None)
        span_height = cell.span_height
        assert span_height == expected_value

    def it_knows_how_many_columns_the_merge_spans(self, width_fixture):
        tc, expected_value = width_fixture
        cell = _Cell(tc, None)
        span_width = cell.span_width
        assert span_width == expected_value

    def it_can_split_a_merged_cell(self, split_fixture):
        origin_tc, range_tcs = split_fixture
        cell = _Cell(origin_tc, None)

        cell.split()

        assert all(tc.gridSpan == 1 for tc in range_tcs)
        assert all(tc.rowSpan == 1 for tc in range_tcs)
        assert all(not tc.hMerge for tc in range_tcs)
        assert all(not tc.vMerge for tc in range_tcs)

    def but_it_raises_when_cell_to_be_split_is_not_merge_origin(self):
        tc = element("a:tbl/a:tr/a:tc").xpath("//a:tc")[0]
        cell = _Cell(tc, None)

        with pytest.raises(ValueError) as e:
            cell.split()
        assert "not a merge-origin cell" in str(e.value)

    def it_knows_what_text_it_contains(self, text_frame_prop_, text_frame_):
        text_frame_prop_.return_value = text_frame_
        text_frame_.text = "foobar"
        cell = _Cell(None, None)

        text = cell.text

        assert text == "foobar"

    def it_can_change_its_text(self, text_frame_prop_, text_frame_):
        text_frame_prop_.return_value = text_frame_
        cell = _Cell(None, None)

        cell.text = "føøbår"

        assert text_frame_.text == "føøbår"

    def it_knows_its_vertical_anchor_setting(self, anchor_get_fixture):
        cell, expected_value = anchor_get_fixture
        assert cell.vertical_anchor == expected_value

    def it_can_change_its_vertical_anchor(self, anchor_set_fixture):
        cell, new_value, expected_xml = anchor_set_fixture
        cell.vertical_anchor = new_value
        assert cell._tc.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("a:tc", None),
            ("a:tc/a:tcPr", None),
            ("a:tc/a:tcPr{anchor=t}", MSO_ANCHOR.TOP),
            ("a:tc/a:tcPr{anchor=ctr}", MSO_ANCHOR.MIDDLE),
            ("a:tc/a:tcPr{anchor=b}", MSO_ANCHOR.BOTTOM),
        ]
    )
    def anchor_get_fixture(self, request):
        tc_cxml, expected_value = request.param
        cell = _Cell(element(tc_cxml), None)
        return cell, expected_value

    @pytest.fixture(
        params=[
            ("a:tc", None, "a:tc"),
            ("a:tc", MSO_ANCHOR.TOP, "a:tc/a:tcPr{anchor=t}"),
            ("a:tc", MSO_ANCHOR.MIDDLE, "a:tc/a:tcPr{anchor=ctr}"),
            ("a:tc", MSO_ANCHOR.BOTTOM, "a:tc/a:tcPr{anchor=b}"),
            ("a:tc/a:tcPr{anchor=t}", MSO_ANCHOR.MIDDLE, "a:tc/a:tcPr{anchor=ctr}"),
            ("a:tc/a:tcPr{anchor=ctr}", None, "a:tc/a:tcPr"),
        ]
    )
    def anchor_set_fixture(self, request):
        tc_cxml, new_value, expected_tc_cxml = request.param
        cell = _Cell(element(tc_cxml), None)
        expected_xml = xml(expected_tc_cxml)
        return cell, new_value, expected_xml

    @pytest.fixture
    def fill_fixture(self, cell):
        return cell

    @pytest.fixture(
        params=[("a:tc", 1), ("a:tc{gridSpan=2}", 1), ("a:tc{rowSpan=42}", 42)]
    )
    def height_fixture(self, request):
        tc_cxml, expected_value = request.param
        tc = element(tc_cxml)
        return tc, expected_value

    @pytest.fixture(
        params=[
            ("a:tc/a:tcPr{marL=82296}", "margin_left", Inches(0.09)),
            ("a:tc/a:tcPr{marR=73152}", "margin_right", Inches(0.08)),
            ("a:tc/a:tcPr{marT=64008}", "margin_top", Inches(0.07)),
            ("a:tc/a:tcPr{marB=54864}", "margin_bottom", Inches(0.06)),
            ("a:tc", "margin_left", Inches(0.1)),
            ("a:tc/a:tcPr", "margin_right", Inches(0.1)),
            ("a:tc", "margin_top", Inches(0.05)),
            ("a:tc/a:tcPr", "margin_bottom", Inches(0.05)),
        ]
    )
    def margin_get_fixture(self, request):
        tc_cxml, margin_prop_name, expected_value = request.param
        cell = _Cell(element(tc_cxml), None)
        return cell, margin_prop_name, expected_value

    @pytest.fixture(
        params=[
            ("a:tc", "margin_left", Inches(0.08), "a:tc/a:tcPr{marL=73152}"),
            ("a:tc", "margin_right", Inches(0.08), "a:tc/a:tcPr{marR=73152}"),
            ("a:tc", "margin_top", Inches(0.08), "a:tc/a:tcPr{marT=73152}"),
            ("a:tc", "margin_bottom", Inches(0.08), "a:tc/a:tcPr{marB=73152}"),
            ("a:tc", "margin_left", None, "a:tc"),
            ("a:tc/a:tcPr{marL=42}", "margin_left", None, "a:tc/a:tcPr"),
        ]
    )
    def margin_set_fixture(self, request):
        tc_cxml, margin_prop_name, new_value, expected_tc_cxml = request.param
        cell = _Cell(element(tc_cxml), None)
        expected_xml = xml(expected_tc_cxml)
        return cell, margin_prop_name, new_value, expected_xml

    @pytest.fixture(
        params=["margin_left", "margin_right", "margin_top", "margin_bottom"]
    )
    def margin_raises_fixture(self, request):
        margin_prop_name = request.param
        cell = _Cell(element("a:tc"), None)
        val_of_invalid_type = "foobar"
        return cell, margin_prop_name, val_of_invalid_type

    @pytest.fixture(
        params=[
            ("a:tc", False),
            ("a:tc{gridSpan=1}", False),
            ("a:tc{hMerge=1}", False),
            ("a:tc{gridSpan=2,vMerge=1}", False),
            ("a:tc{gridSpan=2}", True),
            ("a:tc{rowSpan=2}", True),
            ("a:tc{gridSpan=2,rowSpan=3}", True),
        ]
    )
    def origin_fixture(self, request):
        tc_cxml, expected_value = request.param
        tc = element(tc_cxml)
        return tc, expected_value

    @pytest.fixture(
        params=[
            ("a:tc", False),
            ("a:tc{gridSpan=2}", False),
            ("a:tc{hMerge=1}", True),
            ("a:tc{gridSpan=2,vMerge=1}", True),
            ("a:tc{rowSpan=2,hMerge=true}", True),
            ("a:tc{gridSpan=2,rowSpan=3}", False),
        ]
    )
    def spanned_fixture(self, request):
        tc_cxml, expected_value = request.param
        tc = element(tc_cxml)
        return tc, expected_value

    @pytest.fixture(
        params=[
            (
                "a:tbl/(a:tr/(a:tc{gridSpan=2},a:tc{hMerge=1}),a:tr/(a:tc,a:tc))",
                0,
                [0, 1],
            ),
            (
                "a:tbl/(a:tr/(a:tc{rowSpan=2},a:tc),a:tr/(a:tc{vMerge=1},a:tc))",
                0,
                [0, 2],
            ),
            (
                "a:tbl/(a:tr/(a:tc{gridSpan=2,rowSpan=2},a:tc{hMerge=1,rowSpan=2}),"
                "a:tr/(a:tc{gridSpan=2,vMerge=1},a:tc{hMerge=1,vMerge=1}))",
                0,
                [0, 1, 2, 3],
            ),
        ]
    )
    def split_fixture(self, request):
        tbl_cxml, origin_tc_idx, range_tc_idxs = request.param
        tcs = element(tbl_cxml).xpath("//a:tc")
        origin_tc = tcs[origin_tc_idx]
        range_tcs = tuple(tcs[idx] for idx in range_tc_idxs)
        return origin_tc, range_tcs

    @pytest.fixture(
        params=[("a:tc", 1), ("a:tc{rowSpan=2}", 1), ("a:tc{gridSpan=24}", 24)]
    )
    def width_fixture(self, request):
        tc_cxml, expected_value = request.param
        tc = element(tc_cxml)
        return tc, expected_value

    # fixture components ---------------------------------------------

    @pytest.fixture
    def cell(self):
        return _Cell(element("a:tc"), None)

    @pytest.fixture
    def TcRange_(self, request):
        return class_mock(request, "pptx.table.TcRange")

    @pytest.fixture
    def tc_range_(self, request):
        return instance_mock(request, TcRange)

    @pytest.fixture
    def text_frame_(self, request):
        return instance_mock(request, TextFrame)

    @pytest.fixture
    def text_frame_prop_(self, request):
        return property_mock(request, _Cell, "text_frame")


class Describe_CellCollection(object):
    def it_knows_how_many_cells_it_contains(self, len_fixture):
        cells, expected_count = len_fixture
        assert len(cells) == expected_count

    def it_can_iterate_over_the_cells_it_contains(self, iter_fixture):
        cell_collection, _Cell_, calls, expected_cells = iter_fixture

        cells = list(cell_collection)

        assert _Cell_.call_args_list == calls
        assert cells == expected_cells

    def it_supports_indexed_access(self, _Cell_, cell_):
        tr = element("a:tr/(a:tc, a:tc, a:tc)")
        tcs = tr.xpath("//a:tc")
        _Cell_.return_value = cell_
        cell_collection = _CellCollection(tr, None)

        cell = cell_collection[1]

        _Cell_.assert_called_once_with(tcs[1], cell_collection)
        assert cell is cell_

    def it_raises_on_indexed_access_out_of_range(self):
        cells = _CellCollection(element("a:tr/a:tc"), None)
        with pytest.raises(IndexError):
            cells[-1]
        with pytest.raises(IndexError):
            cells[9]

    # fixtures -------------------------------------------------------

    @pytest.fixture(params=["a:tr", "a:tr/a:tc", "a:tr/(a:tc, a:tc, a:tc)"])
    def iter_fixture(self, request, _Cell_):
        tr_cxml = request.param
        tr = element(tr_cxml)
        tcs = tr.xpath("//a:tc")
        cell_collection = _CellCollection(tr, None)

        expected_cells = [
            instance_mock(request, _Cell, name="cell%d" % idx)
            for idx in range(len(tcs))
        ]
        _Cell_.side_effect = expected_cells
        calls = [call(tc, cell_collection) for tc in tcs]

        return cell_collection, _Cell_, calls, expected_cells

    @pytest.fixture(params=[("a:tr", 0), ("a:tr/a:tc", 1), ("a:tr/(a:tc, a:tc)", 2)])
    def len_fixture(self, request):
        tr_cxml, expected_len = request.param
        cells = _CellCollection(element(tr_cxml), None)
        return cells, expected_len

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _Cell_(self, request):
        return class_mock(request, "pptx.table._Cell")

    @pytest.fixture
    def cell_(self, request):
        return instance_mock(request, _Cell)


class Describe_Column(object):
    def it_knows_its_width(self, width_get_fixture):
        column, expected_value = width_get_fixture
        width = column.width
        assert width == expected_value
        assert isinstance(width, Length)

    def it_can_change_its_width(self, width_set_fixture):
        column, new_width, expected_xml, parent_ = width_set_fixture
        column.width = new_width
        assert column._gridCol.xml == expected_xml
        parent_.notify_width_changed.assert_called_once_with()

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[("a:gridCol{w=914400}", Inches(1)), ("a:gridCol{w=10pt}", Pt(10))]
    )
    def width_get_fixture(self, request):
        gridCol_cxml, expected_value = request.param
        column = _Column(element(gridCol_cxml), None)
        return column, expected_value

    @pytest.fixture(
        params=[
            ("a:gridCol{w=12pt}", Inches(1), "a:gridCol{w=914400}"),
            ("a:gridCol{w=1234}", Inches(1), "a:gridCol{w=914400}"),
        ]
    )
    def width_set_fixture(self, request, parent_):
        gridCol_cxml, new_width, expected_gridCol_cxml = request.param
        column = _Column(element(gridCol_cxml), parent_)
        expected_xml = xml(expected_gridCol_cxml)
        return column, new_width, expected_xml, parent_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def parent_(self, request):
        return instance_mock(request, _ColumnCollection)


class Describe_ColumnCollection(object):
    def it_knows_how_many_columns_it_contains(self, len_fixture):
        columns, expected_count = len_fixture
        assert len(columns) == expected_count

    def it_can_iterate_over_the_columns_it_contains(self, iter_fixture):
        columns, expected_gridCol_lst = iter_fixture
        count = 0
        for idx, column in enumerate(columns):
            assert isinstance(column, _Column)
            assert column._gridCol is expected_gridCol_lst[idx]
            count += 1
        assert count == len(expected_gridCol_lst)

    def it_supports_indexed_access(self, getitem_fixture):
        columns, expected_gridCol_lst = getitem_fixture
        for idx, gridCol in enumerate(expected_gridCol_lst):
            column = columns[idx]
            assert isinstance(column, _Column)
            assert column._gridCol is gridCol

    def it_raises_on_indexed_access_out_of_range(self):
        columns = _ColumnCollection(element("a:tbl/a:tblGrid/a:gridCol"), None)
        with pytest.raises(IndexError):
            columns[-1]
        with pytest.raises(IndexError):
            columns[9]

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            "a:tbl/a:tblGrid",
            "a:tbl/a:tblGrid/a:gridCol",
            "a:tbl/a:tblGrid/(a:gridCol, a:gridCol, a:gridCol)",
        ]
    )
    def getitem_fixture(self, request):
        tbl_cxml = request.param
        tbl = element(tbl_cxml)
        columns = _ColumnCollection(tbl, None)
        expected_column_lst = tbl.xpath("//a:gridCol")
        return columns, expected_column_lst

    @pytest.fixture(
        params=[
            "a:tbl/a:tblGrid",
            "a:tbl/a:tblGrid/a:gridCol",
            "a:tbl/a:tblGrid/(a:gridCol, a:gridCol, a:gridCol)",
        ]
    )
    def iter_fixture(self, request):
        tbl_cxml = request.param
        tbl = element(tbl_cxml)
        columns = _ColumnCollection(tbl, None)
        expected_column_lst = tbl.xpath("//a:gridCol")
        print(expected_column_lst)
        return columns, expected_column_lst

    @pytest.fixture(
        params=[
            ("a:tbl/a:tblGrid", 0),
            ("a:tbl/a:tblGrid/a:gridCol", 1),
            ("a:tbl/a:tblGrid/(a:gridCol,a:gridCol)", 2),
        ]
    )
    def len_fixture(self, request):
        tbl_cxml, expected_len = request.param
        columns = _ColumnCollection(element(tbl_cxml), None)
        return columns, expected_len


class Describe_Row(object):
    def it_knows_its_height(self, height_get_fixture):
        row, expected_value = height_get_fixture
        height = row.height
        assert height == expected_value
        assert isinstance(height, Length)

    def it_can_change_its_height(self, height_set_fixture):
        row, new_height, expected_xml, parent_ = height_set_fixture
        row.height = new_height
        assert row._tr.xml == expected_xml
        parent_.notify_height_changed.assert_called_once_with()

    def it_provides_access_to_its_cells(self, cells_fixture):
        row, _CellCollection_, cells_ = cells_fixture
        cells = row.cells
        _CellCollection_.assert_called_once_with(row._tr, row)
        assert cells is cells_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def cells_fixture(self, _CellCollection_, cells_):
        row = _Row(element("a:tr"), None)
        return row, _CellCollection_, cells_

    @pytest.fixture(params=[("a:tr{h=914400}", Inches(1)), ("a:tr{h=10pt}", Pt(10))])
    def height_get_fixture(self, request):
        tr_cxml, expected_value = request.param
        row = _Row(element(tr_cxml), None)
        return row, expected_value

    @pytest.fixture(
        params=[
            ("a:tr{h=12pt}", Inches(1), "a:tr{h=914400}"),
            ("a:tr{h=1234}", Inches(1), "a:tr{h=914400}"),
        ]
    )
    def height_set_fixture(self, request, parent_):
        tr_cxml, new_height, expected_tr_cxml = request.param
        row = _Row(element(tr_cxml), parent_)
        expected_xml = xml(expected_tr_cxml)
        return row, new_height, expected_xml, parent_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def _CellCollection_(self, request, cells_):
        return class_mock(request, "pptx.table._CellCollection", return_value=cells_)

    @pytest.fixture
    def cells_(self, request):
        return instance_mock(request, _CellCollection)

    @pytest.fixture
    def parent_(self, request):
        return instance_mock(request, _RowCollection)


class Describe_RowCollection(object):
    def it_knows_how_many_rows_it_contains(self, len_fixture):
        rows, expected_count = len_fixture
        assert len(rows) == expected_count

    def it_can_iterate_over_the_rows_it_contains(self, iter_fixture):
        rows, expected_tr_lst = iter_fixture
        count = 0
        for idx, row in enumerate(rows):
            assert isinstance(row, _Row)
            assert row._tr is expected_tr_lst[idx]
            count += 1
        assert count == len(expected_tr_lst)

    def it_supports_indexed_access(self, getitem_fixture):
        rows, expected_tr_lst = getitem_fixture
        for idx, tr in enumerate(expected_tr_lst):
            row = rows[idx]
            assert isinstance(row, _Row)
            assert row._tr is tr

    def it_raises_on_indexed_access_out_of_range(self):
        rows = _RowCollection(element("a:tbl/a:tr"), None)
        with pytest.raises(IndexError):
            rows[-1]
        with pytest.raises(IndexError):
            rows[9]

    # fixtures -------------------------------------------------------

    @pytest.fixture(params=["a:tbl", "a:tbl/a:tr", "a:tbl/(a:tr, a:tr, a:tr)"])
    def getitem_fixture(self, request):
        tbl_cxml = request.param
        tbl = element(tbl_cxml)
        rows = _RowCollection(tbl, None)
        expected_row_lst = tbl.findall(qn("a:tr"))
        return rows, expected_row_lst

    @pytest.fixture(params=["a:tbl", "a:tbl/a:tr", "a:tbl/(a:tr, a:tr, a:tr)"])
    def iter_fixture(self, request):
        tbl_cxml = request.param
        tbl = element(tbl_cxml)
        rows = _RowCollection(tbl, None)
        expected_row_lst = tbl.findall(qn("a:tr"))
        return rows, expected_row_lst

    @pytest.fixture(params=[("a:tbl", 0), ("a:tbl/a:tr", 1), ("a:tbl/(a:tr, a:tr)", 2)])
    def len_fixture(self, request):
        tbl_cxml, expected_len = request.param
        rows = _RowCollection(element(tbl_cxml), None)
        return rows, expected_len



# encoding: utf-8

"""
Axis-related oxml objects.
"""

from __future__ import absolute_import, print_function, unicode_literals

from ...enum.chart import XL_AXIS_CROSSES, XL_TICK_LABEL_POSITION, XL_TICK_MARK
from .shared import CT_Title
from ..simpletypes import ST_AxisUnit, ST_LblOffset
from ..text import CT_TextBody
from ..xmlchemy import (
    BaseOxmlElement,
    OneAndOnlyOne,
    OptionalAttribute,
    RequiredAttribute,
    ZeroOrOne,
)


class BaseAxisElement(BaseOxmlElement):
    """
    Base class for catAx, valAx, and perhaps other axis elements.
    """

    @property
    def defRPr(self):
        """
        ``<a:defRPr>`` great-great-grandchild element, added with its
        ancestors if not present.
        """
        txPr = self.get_or_add_txPr()
        defRPr = txPr.defRPr
        return defRPr

    def _new_title(self):
        return CT_Title.new_title()

    def _new_txPr(self):
        return CT_TextBody.new_txPr()


class CT_AxisUnit(BaseOxmlElement):
    """
    Used for ``<c:majorUnit>`` and ``<c:minorUnit>`` elements, and others.
    """

    val = RequiredAttribute("val", ST_AxisUnit)


class CT_CatAx(BaseAxisElement):
    """
    ``<c:catAx>`` element, defining a category axis.
    """

    _tag_seq = (
        "c:axId",
        "c:scaling",
        "c:delete",
        "c:axPos",
        "c:majorGridlines",
        "c:minorGridlines",
        "c:title",
        "c:numFmt",
        "c:majorTickMark",
        "c:minorTickMark",
        "c:tickLblPos",
        "c:spPr",
        "c:txPr",
        "c:crossAx",
        "c:crosses",
        "c:crossesAt",
        "c:auto",
        "c:lblAlgn",
        "c:lblOffset",
        "c:tickLblSkip",
        "c:tickMarkSkip",
        "c:noMultiLvlLbl",
        "c:extLst",
    )
    scaling = OneAndOnlyOne("c:scaling")
    delete_ = ZeroOrOne("c:delete", successors=_tag_seq[3:])
    majorGridlines = ZeroOrOne("c:majorGridlines", successors=_tag_seq[5:])
    minorGridlines = ZeroOrOne("c:minorGridlines", successors=_tag_seq[6:])
    title = ZeroOrOne("c:title", successors=_tag_seq[7:])
    numFmt = ZeroOrOne("c:numFmt", successors=_tag_seq[8:])
    majorTickMark = ZeroOrOne("c:majorTickMark", successors=_tag_seq[9:])
    minorTickMark = ZeroOrOne("c:minorTickMark", successors=_tag_seq[10:])
    tickLblPos = ZeroOrOne("c:tickLblPos", successors=_tag_seq[11:])
    spPr = ZeroOrOne("c:spPr", successors=_tag_seq[12:])
    txPr = ZeroOrOne("c:txPr", successors=_tag_seq[13:])
    crosses = ZeroOrOne("c:crosses", successors=_tag_seq[15:])
    crossesAt = ZeroOrOne("c:crossesAt", successors=_tag_seq[16:])
    lblOffset = ZeroOrOne("c:lblOffset", successors=_tag_seq[19:])
    del _tag_seq


class CT_ChartLines(BaseOxmlElement):
    """
    Used for c:majorGridlines and c:minorGridlines, specifies gridlines
    visual properties such as color and width.
    """

    spPr = ZeroOrOne("c:spPr", successors=())


class CT_Crosses(BaseAxisElement):
    """
    ``<c:crosses>`` element, specifying where the other axis crosses this
    one.
    """

    val = RequiredAttribute("val", XL_AXIS_CROSSES)


class CT_DateAx(BaseAxisElement):
    """
    ``<c:dateAx>`` element, defining a date (category) axis.
    """

    _tag_seq = (
        "c:axId",
        "c:scaling",
        "c:delete",
        "c:axPos",
        "c:majorGridlines",
        "c:minorGridlines",
        "c:title",
        "c:numFmt",
        "c:majorTickMark",
        "c:minorTickMark",
        "c:tickLblPos",
        "c:spPr",
        "c:txPr",
        "c:crossAx",
        "c:crosses",
        "c:crossesAt",
        "c:auto",
        "c:lblOffset",
        "c:baseTimeUnit",
        "c:majorUnit",
        "c:majorTimeUnit",
        "c:minorUnit",
        "c:minorTimeUnit",
        "c:extLst",
    )
    scaling = OneAndOnlyOne("c:scaling")
    delete_ = ZeroOrOne("c:delete", successors=_tag_seq[3:])
    majorGridlines = ZeroOrOne("c:majorGridlines", successors=_tag_seq[5:])
    minorGridlines = ZeroOrOne("c:minorGridlines", successors=_tag_seq[6:])
    title = ZeroOrOne("c:title", successors=_tag_seq[7:])
    numFmt = ZeroOrOne("c:numFmt", successors=_tag_seq[8:])
    majorTickMark = ZeroOrOne("c:majorTickMark", successors=_tag_seq[9:])
    minorTickMark = ZeroOrOne("c:minorTickMark", successors=_tag_seq[10:])
    tickLblPos = ZeroOrOne("c:tickLblPos", successors=_tag_seq[11:])
    spPr = ZeroOrOne("c:spPr", successors=_tag_seq[12:])
    txPr = ZeroOrOne("c:txPr", successors=_tag_seq[13:])
    crosses = ZeroOrOne("c:crosses", successors=_tag_seq[15:])
    crossesAt = ZeroOrOne("c:crossesAt", successors=_tag_seq[16:])
    lblOffset = ZeroOrOne("c:lblOffset", successors=_tag_seq[18:])
    del _tag_seq


class CT_LblOffset(BaseOxmlElement):
    """
    ``<c:lblOffset>`` custom element class
    """

    val = OptionalAttribute("val", ST_LblOffset, default=100)


class CT_Scaling(BaseOxmlElement):
    """
    ``<c:scaling>`` element, defining axis scale characteristics such as
    maximum value, log vs. linear, etc.
    """

    max = ZeroOrOne("c:max", successors=("c:min", "c:extLst"))
    min = ZeroOrOne("c:min", successors=("c:extLst",))

    @property
    def maximum(self):
        """
        The float value of the ``<c:max>`` child element, or |None| if no max
        element is present.
        """
        max = self.max
        if max is None:
            return None
        return max.val

    @maximum.setter
    def maximum(self, value):
        """
        Set the value of the ``<c:max>`` child element to the float *value*,
        or remove the max element if *value* is |None|.
        """
        self._remove_max()
        if value is None:
            return
        self._add_max(val=value)

    @property
    def minimum(self):
        """
        The float value of the ``<c:min>`` child element, or |None| if no min
        element is present.
        """
        min = self.min
        if min is None:
            return None
        return min.val

    @minimum.setter
    def minimum(self, value):
        """
        Set the value of the ``<c:min>`` child element to the float *value*,
        or remove the min element if *value* is |None|.
        """
        self._remove_min()
        if value is None:
            return
        self._add_min(val=value)


class CT_TickLblPos(BaseOxmlElement):
    """
    ``<c:tickLblPos>`` element.
    """

    val = OptionalAttribute("val", XL_TICK_LABEL_POSITION)


class CT_TickMark(BaseOxmlElement):
    """
    Used for ``<c:minorTickMark>`` and ``<c:majorTickMark>``.
    """

    val = OptionalAttribute("val", XL_TICK_MARK, default=XL_TICK_MARK.CROSS)


class CT_ValAx(BaseAxisElement):
    """
    ``<c:valAx>`` element, defining a value axis.
    """

    _tag_seq = (
        "c:axId",
        "c:scaling",
        "c:delete",
        "c:axPos",
        "c:majorGridlines",
        "c:minorGridlines",
        "c:title",
        "c:numFmt",
        "c:majorTickMark",
        "c:minorTickMark",
        "c:tickLblPos",
        "c:spPr",
        "c:txPr",
        "c:crossAx",
        "c:crosses",
        "c:crossesAt",
        "c:crossBetween",
        "c:majorUnit",
        "c:minorUnit",
        "c:dispUnits",
        "c:extLst",
    )
    scaling = OneAndOnlyOne("c:scaling")
    delete_ = ZeroOrOne("c:delete", successors=_tag_seq[3:])
    majorGridlines = ZeroOrOne("c:majorGridlines", successors=_tag_seq[5:])
    minorGridlines = ZeroOrOne("c:minorGridlines", successors=_tag_seq[6:])
    title = ZeroOrOne("c:title", successors=_tag_seq[7:])
    numFmt = ZeroOrOne("c:numFmt", successors=_tag_seq[8:])
    majorTickMark = ZeroOrOne("c:majorTickMark", successors=_tag_seq[9:])
    minorTickMark = ZeroOrOne("c:minorTickMark", successors=_tag_seq[10:])
    tickLblPos = ZeroOrOne("c:tickLblPos", successors=_tag_seq[11:])
    spPr = ZeroOrOne("c:spPr", successors=_tag_seq[12:])
    txPr = ZeroOrOne("c:txPr", successors=_tag_seq[13:])
    crossAx = ZeroOrOne("c:crossAx", successors=_tag_seq[14:])
    crosses = ZeroOrOne("c:crosses", successors=_tag_seq[15:])
    crossesAt = ZeroOrOne("c:crossesAt", successors=_tag_seq[16:])
    majorUnit = ZeroOrOne("c:majorUnit", successors=_tag_seq[18:])
    minorUnit = ZeroOrOne("c:minorUnit", successors=_tag_seq[19:])
    del _tag_seq



# encoding: utf-8

"""Objects related to construction of freeform shapes."""

from __future__ import absolute_import, division, print_function, unicode_literals

from pptx.compat import Sequence
from pptx.util import lazyproperty


class FreeformBuilder(Sequence):
    """Allows a freeform shape to be specified and created.

    The initial pen position is provided on construction. From there, drawing
    proceeds using successive calls to draw line segments. The freeform shape
    may be closed by calling the :meth:`close` method.

    A shape may have more than one contour, in which case overlapping areas
    are "subtracted". A contour is a sequence of line segments beginning with
    a "move-to" operation. A move-to operation is automatically inserted in
    each new freeform; additional move-to ops can be inserted with the
    `.move_to()` method.
    """

    def __init__(self, shapes, start_x, start_y, x_scale, y_scale):
        super(FreeformBuilder, self).__init__()
        self._shapes = shapes
        self._start_x = start_x
        self._start_y = start_y
        self._x_scale = x_scale
        self._y_scale = y_scale

    def __getitem__(self, idx):
        return self._drawing_operations.__getitem__(idx)

    def __iter__(self):
        return self._drawing_operations.__iter__()

    def __len__(self):
        return self._drawing_operations.__len__()

    @classmethod
    def new(cls, shapes, start_x, start_y, x_scale, y_scale):
        """Return a new |FreeformBuilder| object.

        The initial pen location is specified (in local coordinates) by
        (*start_x*, *start_y*).
        """
        return cls(shapes, int(round(start_x)), int(round(start_y)), x_scale, y_scale)

    def add_line_segments(self, vertices, close=True):
        """Add a straight line segment to each point in *vertices*.

        *vertices* must be an iterable of (x, y) pairs (2-tuples). Each x and
        y value is rounded to the nearest integer before use. The optional
        *close* parameter determines whether the resulting contour is
        *closed* or left *open*.

        Returns this |FreeformBuilder| object so it can be used in chained
        calls.
        """
        for x, y in vertices:
            self._add_line_segment(x, y)
        if close:
            self._add_close()
        return self

    def convert_to_shape(self, origin_x=0, origin_y=0):
        """Return new freeform shape positioned relative to specified offset.

        *origin_x* and *origin_y* locate the origin of the local coordinate
        system in slide coordinates (EMU), perhaps most conveniently by use
        of a |Length| object.

        Note that this method may be called more than once to add multiple
        shapes of the same geometry in different locations on the slide.
        """
        sp = self._add_freeform_sp(origin_x, origin_y)
        path = self._start_path(sp)
        for drawing_operation in self:
            drawing_operation.apply_operation_to(path)
        return self._shapes._shape_factory(sp)

    def move_to(self, x, y):
        """Move pen to (x, y) (local coordinates) without drawing line.

        Returns this |FreeformBuilder| object so it can be used in chained
        calls.
        """
        self._drawing_operations.append(_MoveTo.new(self, x, y))
        return self

    @property
    def shape_offset_x(self):
        """Return x distance of shape origin from local coordinate origin.

        The returned integer represents the leftmost extent of the freeform
        shape, in local coordinates. Note that the bounding box of the shape
        need not start at the local origin.
        """
        min_x = self._start_x
        for drawing_operation in self:
            if hasattr(drawing_operation, "x"):
                min_x = min(min_x, drawing_operation.x)
        return min_x

    @property
    def shape_offset_y(self):
        """Return y distance of shape origin from local coordinate origin.

        The returned integer represents the topmost extent of the freeform
        shape, in local coordinates. Note that the bounding box of the shape
        need not start at the local origin.
        """
        min_y = self._start_y
        for drawing_operation in self:
            if hasattr(drawing_operation, "y"):
                min_y = min(min_y, drawing_operation.y)
        return min_y

    def _add_close(self):
        """Add a close |_Close| operation to the drawing sequence."""
        self._drawing_operations.append(_Close.new())

    def _add_freeform_sp(self, origin_x, origin_y):
        """Add a freeform `p:sp` element having no drawing elements.

        *origin_x* and *origin_y* are specified in slide coordinates, and
        represent the location of the local coordinates origin on the slide.
        """
        spTree = self._shapes._spTree
        return spTree.add_freeform_sp(
            origin_x + self._left, origin_y + self._top, self._width, self._height
        )

    def _add_line_segment(self, x, y):
        """Add a |_LineSegment| operation to the drawing sequence."""
        self._drawing_operations.append(_LineSegment.new(self, x, y))

    @lazyproperty
    def _drawing_operations(self):
        """Return the sequence of drawing operation objects for freeform."""
        return []

    @property
    def _dx(self):
        """Return integer width of this shape's path in local units."""
        min_x = max_x = self._start_x
        for drawing_operation in self:
            if hasattr(drawing_operation, "x"):
                min_x = min(min_x, drawing_operation.x)
                max_x = max(max_x, drawing_operation.x)
        return max_x - min_x

    @property
    def _dy(self):
        """Return integer height of this shape's path in local units."""
        min_y = max_y = self._start_y
        for drawing_operation in self:
            if hasattr(drawing_operation, "y"):
                min_y = min(min_y, drawing_operation.y)
                max_y = max(max_y, drawing_operation.y)
        return max_y - min_y

    @property
    def _height(self):
        """Return vertical size of this shape's path in slide coordinates.

        This value is based on the actual extents of the shape and does not
        include any positioning offset.
        """
        return int(round(self._dy * self._y_scale))

    @property
    def _left(self):
        """Return leftmost extent of this shape's path in slide coordinates.

        Note that this value does not include any positioning offset; it
        assumes the drawing (local) coordinate origin is at (0, 0) on the
        slide.
        """
        return int(round(self.shape_offset_x * self._x_scale))

    def _local_to_shape(self, local_x, local_y):
        """Translate local coordinates point to shape coordinates.

        Shape coordinates have the same unit as local coordinates, but are
        offset such that the origin of the shape coordinate system (0, 0) is
        located at the top-left corner of the shape bounding box.
        """
        return (local_x - self.shape_offset_x, local_y - self.shape_offset_y)

    def _start_path(self, sp):
        """Return a newly created `a:path` element added to *sp*.

        The returned `a:path` element has an `a:moveTo` element representing
        the shape starting point as its only child.
        """
        path = sp.add_path(w=self._dx, h=self._dy)
        path.add_moveTo(*self._local_to_shape(self._start_x, self._start_y))
        return path

    @property
    def _top(self):
        """Return topmost extent of this shape's path in slide coordinates.

        Note that this value does not include any positioning offset; it
        assumes the drawing (local) coordinate origin is located at slide
        coordinates (0, 0) (top-left corner of slide).
        """
        return int(round(self.shape_offset_y * self._y_scale))

    @property
    def _width(self):
        """Return width of this shape's path in slide coordinates.

        This value is based on the actual extents of the shape path and does
        not include any positioning offset.
        """
        return int(round(self._dx * self._x_scale))


class _BaseDrawingOperation(object):
    """Base class for freeform drawing operations.

    A drawing operation has at least one location (x, y) in local
    coordinates.
    """

    def __init__(self, freeform_builder, x, y):
        super(_BaseDrawingOperation, self).__init__()
        self._freeform_builder = freeform_builder
        self._x = x
        self._y = y

    def apply_operation_to(self, path):
        """Add the XML element(s) implementing this operation to *path*.

        Must be implemented by each subclass.
        """
        raise NotImplementedError("must be implemented by each subclass")

    @property
    def x(self):
        """Return the horizontal (x) target location of this operation.

        The returned value is an integer in local coordinates.
        """
        return self._x

    @property
    def y(self):
        """Return the vertical (y) target location of this operation.

        The returned value is an integer in local coordinates.
        """
        return self._y


class _Close(object):
    """Specifies adding a `<a:close/>` element to the current contour."""

    @classmethod
    def new(cls):
        """Return a new _Close object."""
        return cls()

    def apply_operation_to(self, path):
        """Add `a:close` element to *path*."""
        return path.add_close()


class _LineSegment(_BaseDrawingOperation):
    """Specifies a straight line segment ending at the specified point."""

    @classmethod
    def new(cls, freeform_builder, x, y):
        """Return a new _LineSegment object ending at point *(x, y)*.

        Both *x* and *y* are rounded to the nearest integer before use.
        """
        return cls(freeform_builder, int(round(x)), int(round(y)))

    def apply_operation_to(self, path):
        """Add `a:lnTo` element to *path* for this line segment.

        Returns the `a:lnTo` element newly added to the path.
        """
        return path.add_lnTo(
            self._x - self._freeform_builder.shape_offset_x,
            self._y - self._freeform_builder.shape_offset_y,
        )


class _MoveTo(_BaseDrawingOperation):
    """Specifies a new pen position."""

    @classmethod
    def new(cls, freeform_builder, x, y):
        """Return a new _MoveTo object for move to point *(x, y)*.

        Both *x* and *y* are rounded to the nearest integer before use.
        """
        return cls(freeform_builder, int(round(x)), int(round(y)))

    def apply_operation_to(self, path):
        """Add `a:moveTo` element to *path* for this line segment."""
        return path.add_moveTo(
            self._x - self._freeform_builder.shape_offset_x,
            self._y - self._freeform_builder.shape_offset_y,
        )



# encoding: utf-8

"""
Objects used across sub-package
"""


class Subshape(object):
    """
    Provides common services for drawing elements that occur below a shape
    but may occasionally require an ancestor object to provide a service,
    such as add or drop a relationship. Provides ``self._parent`` attribute
    to subclasses.
    """

    def __init__(self, parent):
        super(Subshape, self).__init__()
        self._parent = parent

    @property
    def part(self):
        """
        The package part containing this object
        """
        return self._parent.part



# encoding: utf-8

"""
Test suite for pptx.oxml.__init__.py module, primarily XML parser-related.
"""

from __future__ import print_function, unicode_literals

import pytest

from lxml import etree

from pptx.oxml import oxml_parser, parse_xml, register_element_cls
from pptx.oxml.ns import qn
from pptx.oxml.xmlchemy import BaseOxmlElement

from ..unitutil.mock import function_mock, loose_mock, var_mock


class DescribeOxmlParser(object):
    def it_strips_whitespace_between_elements(self, foo, stripped_xml_bytes):
        xml_bytes = etree.tostring(foo)
        assert xml_bytes == stripped_xml_bytes


class DescribeParseXml(object):
    def it_uses_oxml_configured_parser_to_parse_xml(
        self, mock_xml_bytes, fromstring, mock_oxml_parser
    ):
        element = parse_xml(mock_xml_bytes)
        fromstring.assert_called_once_with(mock_xml_bytes, mock_oxml_parser)
        assert element is fromstring.return_value

    def it_prefers_to_parse_bytes(self, xml_bytes):
        parse_xml(xml_bytes)

    def but_accepts_unicode_providing_there_is_no_encoding_declaration(self):
        non_enc_decl = '<?xml version="1.0" standalone="yes"?>'
        enc_decl = '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>'
        xml_body = "<foo><bar>føøbår</bar></foo>"
        # unicode body by itself doesn't raise
        parse_xml(xml_body)
        # adding XML decl without encoding attr doesn't raise either
        xml_text = "%s\n%s" % (non_enc_decl, xml_body)
        parse_xml(xml_text)
        # but adding encoding in the declaration raises ValueError
        xml_text = "%s\n%s" % (enc_decl, xml_body)
        with pytest.raises(ValueError):
            parse_xml(xml_text)


class DescribeRegisterCustomElementClass(object):
    def it_determines_cust_elm_class_constructed_for_specified_tag(self, xml_bytes):
        register_element_cls("a:foo", CustElmCls)
        foo = etree.fromstring(xml_bytes, oxml_parser)
        assert type(foo) is CustElmCls
        assert type(foo.find(qn("a:bar"))) is etree._Element


# ===========================================================================
# fixtures
# ===========================================================================


class CustElmCls(BaseOxmlElement):
    pass


@pytest.fixture
def foo(xml_bytes):
    return etree.fromstring(xml_bytes, oxml_parser)


@pytest.fixture
def fromstring(request):
    return function_mock(request, "pptx.oxml.etree.fromstring")


@pytest.fixture
def mock_oxml_parser(request):
    return var_mock(request, "pptx.oxml.oxml_parser")


@pytest.fixture
def mock_xml_bytes(request):
    return loose_mock(request, "xml_bytes")


@pytest.fixture
def stripped_xml_bytes():
    return (
        '<a:foo xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/ma'
        'in"><a:bar>foobar</a:bar></a:foo>'
    ).encode("utf-8")


@pytest.fixture
def xml_bytes(xml_text):
    return xml_text.encode("utf-8")


@pytest.fixture
def xml_text():
    return (
        '<?xml version="1.0" encoding="UTF-8" standalone="yes"?>\n'
        '<a:foo xmlns:a="http://schemas.openxmlformats.org/drawingml/2006/ma'
        'in">\n'
        "  <a:bar>foobar</a:bar>\n"
        "</a:foo>\n"
    )



# encoding: utf-8

"""
Test suite for pptx.chart module
"""

from __future__ import absolute_import, print_function

import pytest

from pptx.chart.axis import (
    AxisTitle,
    _BaseAxis,
    CategoryAxis,
    DateAxis,
    MajorGridlines,
    TickLabels,
    ValueAxis,
)
from pptx.dml.chtfmt import ChartFormat
from pptx.enum.chart import (
    XL_AXIS_CROSSES,
    XL_CATEGORY_TYPE,
    XL_TICK_LABEL_POSITION as XL_TICK_LBL_POS,
    XL_TICK_MARK,
)
from pptx.text.text import Font

from ..unitutil.cxml import element, xml
from ..unitutil.mock import class_mock, instance_mock


class Describe_BaseAxis(object):
    def it_knows_whether_it_has_major_gridlines(self, major_gridlines_get_fixture):
        base_axis, expected_value = major_gridlines_get_fixture
        assert base_axis.has_major_gridlines is expected_value

    def it_can_change_whether_it_has_major_gridlines(self, major_gridlines_set_fixture):
        base_axis, new_value, expected_xml = major_gridlines_set_fixture
        base_axis.has_major_gridlines = new_value
        assert base_axis._element.xml == expected_xml

    def it_knows_whether_it_has_minor_gridlines(self, minor_gridlines_get_fixture):
        base_axis, expected_value = minor_gridlines_get_fixture
        assert base_axis.has_minor_gridlines is expected_value

    def it_can_change_whether_it_has_minor_gridlines(self, minor_gridlines_set_fixture):
        base_axis, new_value, expected_xml = minor_gridlines_set_fixture
        base_axis.has_minor_gridlines = new_value
        assert base_axis._element.xml == expected_xml

    def it_knows_whether_it_has_a_title(self, has_title_get_fixture):
        axis, expected_value = has_title_get_fixture
        assert axis.has_title is expected_value

    def it_can_change_whether_it_has_a_title(self, has_title_set_fixture):
        axis, new_value, expected_xml = has_title_set_fixture
        axis.has_title = new_value
        assert axis._element.xml == expected_xml

    def it_knows_whether_it_is_visible(self, visible_get_fixture):
        axis, expected_bool_value = visible_get_fixture
        assert axis.visible is expected_bool_value

    def it_can_change_whether_it_is_visible(self, visible_set_fixture):
        axis, new_value, expected_xml = visible_set_fixture
        axis.visible = new_value
        assert axis._element.xml == expected_xml

    def it_raises_on_assign_non_bool_to_visible(self):
        axis = _BaseAxis(None)
        with pytest.raises(ValueError):
            axis.visible = "foobar"

    def it_knows_the_scale_maximum(self, maximum_scale_get_fixture):
        axis, expected_value = maximum_scale_get_fixture
        assert axis.maximum_scale == expected_value

    def it_can_change_the_scale_maximum(self, maximum_scale_set_fixture):
        axis, new_value, expected_xml = maximum_scale_set_fixture
        axis.maximum_scale = new_value
        assert axis._element.xml == expected_xml

    def it_knows_the_scale_minimum(self, minimum_scale_get_fixture):
        axis, expected_value = minimum_scale_get_fixture
        assert axis.minimum_scale == expected_value

    def it_can_change_the_scale_minimum(self, minimum_scale_set_fixture):
        axis, new_value, expected_xml = minimum_scale_set_fixture
        axis.minimum_scale = new_value
        assert axis._element.xml == expected_xml

    def it_knows_its_major_tick_setting(self, major_tick_get_fixture):
        axis, expected_value = major_tick_get_fixture
        assert axis.major_tick_mark == expected_value

    def it_can_change_its_major_tick_mark(self, major_tick_set_fixture):
        axis, new_value, expected_xml = major_tick_set_fixture
        axis.major_tick_mark = new_value
        assert axis._element.xml == expected_xml

    def it_knows_its_minor_tick_setting(self, minor_tick_get_fixture):
        axis, expected_value = minor_tick_get_fixture
        assert axis.minor_tick_mark == expected_value

    def it_can_change_its_minor_tick_mark(self, minor_tick_set_fixture):
        axis, new_value, expected_xml = minor_tick_set_fixture
        axis.minor_tick_mark = new_value
        assert axis._element.xml == expected_xml

    def it_knows_its_tick_label_position(self, tick_lbl_pos_get_fixture):
        axis, expected_value = tick_lbl_pos_get_fixture
        assert axis.tick_label_position == expected_value

    def it_can_change_its_tick_label_position(self, tick_lbl_pos_set_fixture):
        axis, new_value, expected_xml = tick_lbl_pos_set_fixture
        axis.tick_label_position = new_value
        assert axis._element.xml == expected_xml

    def it_provides_access_to_its_title(self, title_fixture):
        axis, AxisTitle_, axis_title_ = title_fixture
        axis_title = axis.axis_title
        AxisTitle_.assert_called_once_with(axis._element.title)
        assert axis_title is axis_title_

    def it_provides_access_to_its_format(self, format_fixture):
        axis, ChartFormat_, format_ = format_fixture
        format = axis.format
        ChartFormat_.assert_called_once_with(axis._xAx)
        assert format is format_

    def it_provides_access_to_its_major_gridlines(self, maj_grdlns_fixture):
        axis, MajorGridlines_, xAx, major_gridlines_ = maj_grdlns_fixture
        major_gridlines = axis.major_gridlines
        MajorGridlines_.assert_called_once_with(xAx)
        assert major_gridlines is major_gridlines_

    def it_provides_access_to_the_tick_labels(self, tick_labels_fixture):
        axis, tick_labels_, TickLabels_, xAx = tick_labels_fixture
        tick_labels = axis.tick_labels
        TickLabels_.assert_called_once_with(xAx)
        assert tick_labels is tick_labels_

    # fixtures -------------------------------------------------------

    @pytest.fixture(params=["c:catAx", "c:dateAx", "c:valAx"])
    def format_fixture(self, request, ChartFormat_, format_):
        xAx_cxml = request.param
        axis = _BaseAxis(element(xAx_cxml))
        return axis, ChartFormat_, format_

    @pytest.fixture(
        params=[
            ("c:catAx", False),
            ("c:catAx/c:title", True),
            ("c:dateAx", False),
            ("c:dateAx/c:title", True),
            ("c:valAx", False),
            ("c:valAx/c:title", True),
        ]
    )
    def has_title_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        axis = _BaseAxis(element(xAx_cxml))
        return axis, expected_value

    @pytest.fixture(
        params=[
            ("c:catAx", True, "c:catAx/c:title/(c:layout,c:overlay{val=0})"),
            ("c:catAx/c:title", True, "c:catAx/c:title"),
            ("c:catAx/c:title", False, "c:catAx"),
            ("c:catAx", False, "c:catAx"),
            ("c:dateAx", True, "c:dateAx/c:title/(c:layout,c:overlay{val=0})"),
            ("c:dateAx/c:title", True, "c:dateAx/c:title"),
            ("c:dateAx/c:title", False, "c:dateAx"),
            ("c:dateAx", False, "c:dateAx"),
            ("c:valAx", True, "c:valAx/c:title/(c:layout,c:overlay{val=0})"),
            ("c:valAx/c:title", True, "c:valAx/c:title"),
            ("c:valAx/c:title", False, "c:valAx"),
            ("c:valAx", False, "c:valAx"),
        ]
    )
    def has_title_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        axis = _BaseAxis(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return axis, new_value, expected_xml

    @pytest.fixture(params=["c:catAx", "c:dateAx", "c:valAx"])
    def maj_grdlns_fixture(self, request, MajorGridlines_, major_gridlines_):
        xAx_cxml = request.param
        xAx = element(xAx_cxml)
        axis = _BaseAxis(xAx)
        return axis, MajorGridlines_, xAx, major_gridlines_

    @pytest.fixture(
        params=[
            ("c:catAx", False),
            ("c:catAx/c:majorGridlines", True),
            ("c:dateAx", False),
            ("c:dateAx/c:majorGridlines", True),
            ("c:valAx", False),
            ("c:valAx/c:majorGridlines", True),
        ]
    )
    def major_gridlines_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        base_axis = _BaseAxis(element(xAx_cxml))
        return base_axis, expected_value

    @pytest.fixture(
        params=[
            ("c:catAx", True, "c:catAx/c:majorGridlines"),
            ("c:catAx/c:majorGridlines", True, "c:catAx/c:majorGridlines"),
            ("c:catAx/c:majorGridlines", False, "c:catAx"),
            ("c:catAx", False, "c:catAx"),
            ("c:dateAx", True, "c:dateAx/c:majorGridlines"),
            ("c:dateAx/c:majorGridlines", True, "c:dateAx/c:majorGridlines"),
            ("c:dateAx/c:majorGridlines", False, "c:dateAx"),
            ("c:dateAx", False, "c:dateAx"),
            ("c:valAx", True, "c:valAx/c:majorGridlines"),
            ("c:valAx/c:majorGridlines", True, "c:valAx/c:majorGridlines"),
            ("c:valAx/c:majorGridlines", False, "c:valAx"),
            ("c:valAx", False, "c:valAx"),
        ]
    )
    def major_gridlines_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        base_axis = _BaseAxis(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return base_axis, new_value, expected_xml

    @pytest.fixture(
        params=[
            ("c:catAx", XL_TICK_MARK.CROSS),
            ("c:catAx/c:majorTickMark", XL_TICK_MARK.CROSS),
            ("c:catAx/c:majorTickMark{val=out}", XL_TICK_MARK.OUTSIDE),
            ("c:dateAx", XL_TICK_MARK.CROSS),
            ("c:dateAx/c:majorTickMark", XL_TICK_MARK.CROSS),
            ("c:dateAx/c:majorTickMark{val=out}", XL_TICK_MARK.OUTSIDE),
            ("c:valAx", XL_TICK_MARK.CROSS),
            ("c:valAx/c:majorTickMark", XL_TICK_MARK.CROSS),
            ("c:valAx/c:majorTickMark{val=in}", XL_TICK_MARK.INSIDE),
        ]
    )
    def major_tick_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        axis = _BaseAxis(element(xAx_cxml))
        return axis, expected_value

    @pytest.fixture(
        params=[
            ("c:catAx", XL_TICK_MARK.INSIDE, "c:catAx/c:majorTickMark{val=in}"),
            (
                "c:catAx/c:majorTickMark{val=in}",
                XL_TICK_MARK.OUTSIDE,
                "c:catAx/c:majorTickMark{val=out}",
            ),
            ("c:catAx/c:majorTickMark{val=out}", XL_TICK_MARK.CROSS, "c:catAx"),
            ("c:catAx", XL_TICK_MARK.CROSS, "c:catAx"),
            ("c:catAx/c:majorTickMark{val=cross}", XL_TICK_MARK.CROSS, "c:catAx"),
            ("c:dateAx", XL_TICK_MARK.INSIDE, "c:dateAx/c:majorTickMark{val=in}"),
            (
                "c:dateAx/c:majorTickMark{val=in}",
                XL_TICK_MARK.OUTSIDE,
                "c:dateAx/c:majorTickMark{val=out}",
            ),
            ("c:dateAx/c:majorTickMark{val=out}", XL_TICK_MARK.CROSS, "c:dateAx"),
            ("c:dateAx", XL_TICK_MARK.CROSS, "c:dateAx"),
            ("c:dateAx/c:majorTickMark{val=cross}", XL_TICK_MARK.CROSS, "c:dateAx"),
            ("c:valAx", XL_TICK_MARK.INSIDE, "c:valAx/c:majorTickMark{val=in}"),
            (
                "c:valAx/c:majorTickMark{val=in}",
                XL_TICK_MARK.OUTSIDE,
                "c:valAx/c:majorTickMark{val=out}",
            ),
            ("c:valAx/c:majorTickMark{val=out}", XL_TICK_MARK.CROSS, "c:valAx"),
            ("c:valAx", XL_TICK_MARK.CROSS, "c:valAx"),
            ("c:valAx/c:majorTickMark{val=cross}", XL_TICK_MARK.CROSS, "c:valAx"),
        ]
    )
    def major_tick_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        axis = _BaseAxis(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return axis, new_value, expected_xml

    @pytest.fixture(
        params=[
            ("c:catAx/c:scaling", None),
            ("c:catAx/c:scaling/c:max{val=12.34}", 12.34),
            ("c:dateAx/c:scaling", None),
            ("c:dateAx/c:scaling/c:max{val=42.24}", 42.24),
            ("c:valAx/c:scaling", None),
            ("c:valAx/c:scaling/c:max{val=23.45}", 23.45),
        ]
    )
    def maximum_scale_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        axis = _BaseAxis(element(xAx_cxml))
        return axis, expected_value

    @pytest.fixture(
        params=[
            ("c:catAx/c:scaling", 34.56, "c:catAx/c:scaling/c:max{val=34.56}"),
            (
                "c:catAx/c:scaling/c:max{val=34.56}",
                42.42,
                "c:catAx/c:scaling/c:max{val=42.42}",
            ),
            ("c:catAx/c:scaling/c:max{val=42.42}", None, "c:catAx/c:scaling"),
            ("c:catAx/c:scaling", None, "c:catAx/c:scaling"),
            ("c:dateAx/c:scaling", 45.67, "c:dateAx/c:scaling/c:max{val=45.67}"),
            (
                "c:dateAx/c:scaling/c:max{val=45.67}",
                42.42,
                "c:dateAx/c:scaling/c:max{val=42.42}",
            ),
            ("c:dateAx/c:scaling/c:max{val=42.42}", None, "c:dateAx/c:scaling"),
            ("c:dateAx/c:scaling", None, "c:dateAx/c:scaling"),
            ("c:valAx/c:scaling", 56.78, "c:valAx/c:scaling/c:max{val=56.78}"),
            (
                "c:valAx/c:scaling/c:max{val=56.78}",
                42.42,
                "c:valAx/c:scaling/c:max{val=42.42}",
            ),
            ("c:valAx/c:scaling/c:max{val=42.42}", None, "c:valAx/c:scaling"),
            ("c:valAx/c:scaling", None, "c:valAx/c:scaling"),
        ]
    )
    def maximum_scale_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        axis = _BaseAxis(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return axis, new_value, expected_xml

    @pytest.fixture(
        params=[
            ("c:catAx/c:scaling", None),
            ("c:catAx/c:scaling/c:min{val=12.34}", 12.34),
            ("c:dateAx/c:scaling", None),
            ("c:dateAx/c:scaling/c:min{val=42.24}", 42.24),
            ("c:valAx/c:scaling", None),
            ("c:valAx/c:scaling/c:min{val=23.45}", 23.45),
        ]
    )
    def minimum_scale_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        axis = _BaseAxis(element(xAx_cxml))
        return axis, expected_value

    @pytest.fixture(
        params=[
            ("c:catAx/c:scaling", 34.56, "c:catAx/c:scaling/c:min{val=34.56}"),
            (
                "c:catAx/c:scaling/c:min{val=34.56}",
                42.42,
                "c:catAx/c:scaling/c:min{val=42.42}",
            ),
            ("c:catAx/c:scaling/c:min{val=42.42}", None, "c:catAx/c:scaling"),
            ("c:catAx/c:scaling", None, "c:catAx/c:scaling"),
            ("c:dateAx/c:scaling", 45.67, "c:dateAx/c:scaling/c:min{val=45.67}"),
            (
                "c:dateAx/c:scaling/c:min{val=45.67}",
                42.42,
                "c:dateAx/c:scaling/c:min{val=42.42}",
            ),
            ("c:dateAx/c:scaling/c:min{val=42.42}", None, "c:dateAx/c:scaling"),
            ("c:dateAx/c:scaling", None, "c:dateAx/c:scaling"),
            ("c:valAx/c:scaling", 56.78, "c:valAx/c:scaling/c:min{val=56.78}"),
            (
                "c:valAx/c:scaling/c:min{val=56.78}",
                42.42,
                "c:valAx/c:scaling/c:min{val=42.42}",
            ),
            ("c:valAx/c:scaling/c:min{val=42.42}", None, "c:valAx/c:scaling"),
            ("c:valAx/c:scaling", None, "c:valAx/c:scaling"),
        ]
    )
    def minimum_scale_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        axis = _BaseAxis(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return axis, new_value, expected_xml

    @pytest.fixture(
        params=[
            ("c:catAx", False),
            ("c:catAx/c:minorGridlines", True),
            ("c:dateAx", False),
            ("c:dateAx/c:minorGridlines", True),
            ("c:valAx", False),
            ("c:valAx/c:minorGridlines", True),
        ]
    )
    def minor_gridlines_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        base_axis = _BaseAxis(element(xAx_cxml))
        return base_axis, expected_value

    @pytest.fixture(
        params=[
            ("c:catAx", True, "c:catAx/c:minorGridlines"),
            ("c:catAx/c:minorGridlines", True, "c:catAx/c:minorGridlines"),
            ("c:catAx/c:minorGridlines", False, "c:catAx"),
            ("c:catAx", False, "c:catAx"),
            ("c:dateAx", True, "c:dateAx/c:minorGridlines"),
            ("c:dateAx/c:minorGridlines", True, "c:dateAx/c:minorGridlines"),
            ("c:dateAx/c:minorGridlines", False, "c:dateAx"),
            ("c:dateAx", False, "c:dateAx"),
            ("c:valAx", True, "c:valAx/c:minorGridlines"),
            ("c:valAx/c:minorGridlines", True, "c:valAx/c:minorGridlines"),
            ("c:valAx/c:minorGridlines", False, "c:valAx"),
            ("c:valAx", False, "c:valAx"),
        ]
    )
    def minor_gridlines_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        base_axis = _BaseAxis(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return base_axis, new_value, expected_xml

    @pytest.fixture(
        params=[
            ("c:catAx", XL_TICK_MARK.CROSS),
            ("c:catAx/c:minorTickMark", XL_TICK_MARK.CROSS),
            ("c:catAx/c:minorTickMark{val=out}", XL_TICK_MARK.OUTSIDE),
            ("c:dateAx", XL_TICK_MARK.CROSS),
            ("c:dateAx/c:minorTickMark", XL_TICK_MARK.CROSS),
            ("c:dateAx/c:minorTickMark{val=out}", XL_TICK_MARK.OUTSIDE),
            ("c:valAx", XL_TICK_MARK.CROSS),
            ("c:valAx/c:minorTickMark", XL_TICK_MARK.CROSS),
            ("c:valAx/c:minorTickMark{val=in}", XL_TICK_MARK.INSIDE),
        ]
    )
    def minor_tick_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        axis = _BaseAxis(element(xAx_cxml))
        return axis, expected_value

    @pytest.fixture(
        params=[
            ("c:catAx", XL_TICK_MARK.INSIDE, "c:catAx/c:minorTickMark{val=in}"),
            (
                "c:catAx/c:minorTickMark{val=in}",
                XL_TICK_MARK.OUTSIDE,
                "c:catAx/c:minorTickMark{val=out}",
            ),
            ("c:catAx/c:minorTickMark{val=out}", XL_TICK_MARK.CROSS, "c:catAx"),
            ("c:catAx", XL_TICK_MARK.CROSS, "c:catAx"),
            ("c:catAx/c:minorTickMark{val=cross}", XL_TICK_MARK.CROSS, "c:catAx"),
            ("c:dateAx", XL_TICK_MARK.INSIDE, "c:dateAx/c:minorTickMark{val=in}"),
            (
                "c:dateAx/c:minorTickMark{val=in}",
                XL_TICK_MARK.OUTSIDE,
                "c:dateAx/c:minorTickMark{val=out}",
            ),
            ("c:dateAx/c:minorTickMark{val=out}", XL_TICK_MARK.CROSS, "c:dateAx"),
            ("c:dateAx", XL_TICK_MARK.CROSS, "c:dateAx"),
            ("c:dateAx/c:minorTickMark{val=cross}", XL_TICK_MARK.CROSS, "c:dateAx"),
            ("c:valAx", XL_TICK_MARK.INSIDE, "c:valAx/c:minorTickMark{val=in}"),
            (
                "c:valAx/c:minorTickMark{val=in}",
                XL_TICK_MARK.OUTSIDE,
                "c:valAx/c:minorTickMark{val=out}",
            ),
            ("c:valAx/c:minorTickMark{val=out}", XL_TICK_MARK.CROSS, "c:valAx"),
            ("c:valAx", XL_TICK_MARK.CROSS, "c:valAx"),
            ("c:valAx/c:minorTickMark{val=cross}", XL_TICK_MARK.CROSS, "c:valAx"),
        ]
    )
    def minor_tick_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        axis = _BaseAxis(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return axis, new_value, expected_xml

    @pytest.fixture(params=["c:catAx", "c:dateAx", "c:valAx"])
    def tick_labels_fixture(self, request, TickLabels_, tick_labels_):
        xAx_cxml = request.param
        xAx = element(xAx_cxml)
        axis = _BaseAxis(xAx)
        return axis, tick_labels_, TickLabels_, xAx

    @pytest.fixture(
        params=[
            ("c:catAx", XL_TICK_LBL_POS.NEXT_TO_AXIS),
            ("c:catAx/c:tickLblPos", XL_TICK_LBL_POS.NEXT_TO_AXIS),
            ("c:catAx/c:tickLblPos{val=high}", XL_TICK_LBL_POS.HIGH),
            ("c:dateAx", XL_TICK_LBL_POS.NEXT_TO_AXIS),
            ("c:dateAx/c:tickLblPos", XL_TICK_LBL_POS.NEXT_TO_AXIS),
            ("c:dateAx/c:tickLblPos{val=low}", XL_TICK_LBL_POS.LOW),
            ("c:valAx", XL_TICK_LBL_POS.NEXT_TO_AXIS),
            ("c:valAx/c:tickLblPos", XL_TICK_LBL_POS.NEXT_TO_AXIS),
            ("c:valAx/c:tickLblPos{val=none}", XL_TICK_LBL_POS.NONE),
        ]
    )
    def tick_lbl_pos_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        axis = _BaseAxis(element(xAx_cxml))
        return axis, expected_value

    @pytest.fixture(
        params=[
            ("c:catAx", XL_TICK_LBL_POS.HIGH, "c:catAx/c:tickLblPos{val=high}"),
            (
                "c:catAx/c:tickLblPos{val=high}",
                XL_TICK_LBL_POS.LOW,
                "c:catAx/c:tickLblPos{val=low}",
            ),
            ("c:catAx/c:tickLblPos{val=low}", None, "c:catAx/c:tickLblPos"),
            ("c:catAx", None, "c:catAx/c:tickLblPos"),
            (
                "c:dateAx",
                XL_TICK_LBL_POS.NEXT_TO_AXIS,
                "c:dateAx/c:tickLblPos{val=nextTo}",
            ),
            (
                "c:dateAx/c:tickLblPos{val=nextTo}",
                XL_TICK_LBL_POS.NONE,
                "c:dateAx/c:tickLblPos{val=none}",
            ),
            ("c:dateAx/c:tickLblPos{val=none}", None, "c:dateAx/c:tickLblPos"),
            ("c:valAx", XL_TICK_LBL_POS.HIGH, "c:valAx/c:tickLblPos{val=high}"),
            (
                "c:valAx/c:tickLblPos{val=high}",
                XL_TICK_LBL_POS.LOW,
                "c:valAx/c:tickLblPos{val=low}",
            ),
            ("c:valAx/c:tickLblPos{val=low}", None, "c:valAx/c:tickLblPos"),
        ]
    )
    def tick_lbl_pos_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        axis = _BaseAxis(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return axis, new_value, expected_xml

    @pytest.fixture(params=["c:catAx", "c:dateAx", "c:valAx"])
    def title_fixture(self, request, AxisTitle_, axis_title_):
        xAx_cxml = request.param
        axis = _BaseAxis(element(xAx_cxml))
        return axis, AxisTitle_, axis_title_

    @pytest.fixture(
        params=[
            ("c:catAx", False),
            ("c:catAx/c:delete", False),
            ("c:catAx/c:delete{val=0}", True),
            ("c:catAx/c:delete{val=1}", False),
            ("c:catAx/c:delete{val=false}", True),
            ("c:dateAx", False),
            ("c:dateAx/c:delete", False),
            ("c:dateAx/c:delete{val=0}", True),
            ("c:dateAx/c:delete{val=1}", False),
            ("c:dateAx/c:delete{val=false}", True),
            ("c:valAx", False),
            ("c:valAx/c:delete", False),
            ("c:valAx/c:delete{val=0}", True),
            ("c:valAx/c:delete{val=1}", False),
            ("c:valAx/c:delete{val=false}", True),
        ]
    )
    def visible_get_fixture(self, request):
        xAx_cxml, expected_bool_value = request.param
        axis = _BaseAxis(element(xAx_cxml))
        return axis, expected_bool_value

    @pytest.fixture(
        params=[
            ("c:catAx", False, "c:catAx/c:delete"),
            ("c:catAx/c:delete", True, "c:catAx/c:delete{val=0}"),
            ("c:catAx/c:delete{val=1}", True, "c:catAx/c:delete{val=0}"),
            ("c:catAx/c:delete{val=0}", False, "c:catAx/c:delete"),
            ("c:catAx", True, "c:catAx/c:delete{val=0}"),
            ("c:dateAx", False, "c:dateAx/c:delete"),
            ("c:dateAx/c:delete", True, "c:dateAx/c:delete{val=0}"),
            ("c:dateAx/c:delete{val=0}", False, "c:dateAx/c:delete"),
            ("c:dateAx", True, "c:dateAx/c:delete{val=0}"),
            ("c:valAx/c:delete", True, "c:valAx/c:delete{val=0}"),
            ("c:valAx/c:delete{val=1}", False, "c:valAx/c:delete"),
            ("c:valAx/c:delete{val=0}", True, "c:valAx/c:delete{val=0}"),
        ]
    )
    def visible_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        axis = _BaseAxis(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return axis, new_value, expected_xml

    # fixture components ---------------------------------------------

    @pytest.fixture
    def AxisTitle_(self, request, axis_title_):
        return class_mock(
            request, "pptx.chart.axis.AxisTitle", return_value=axis_title_
        )

    @pytest.fixture
    def axis_title_(self, request):
        return instance_mock(request, AxisTitle)

    @pytest.fixture
    def ChartFormat_(self, request, format_):
        return class_mock(request, "pptx.chart.axis.ChartFormat", return_value=format_)

    @pytest.fixture
    def format_(self, request):
        return instance_mock(request, ChartFormat)

    @pytest.fixture
    def MajorGridlines_(self, request, major_gridlines_):
        return class_mock(
            request, "pptx.chart.axis.MajorGridlines", return_value=major_gridlines_
        )

    @pytest.fixture
    def major_gridlines_(self, request):
        return instance_mock(request, MajorGridlines)

    @pytest.fixture
    def TickLabels_(self, request, tick_labels_):
        return class_mock(
            request, "pptx.chart.axis.TickLabels", return_value=tick_labels_
        )

    @pytest.fixture
    def tick_labels_(self, request):
        return instance_mock(request, TickLabels)


class DescribeAxisTitle(object):
    def it_knows_whether_it_has_a_text_frame(self, has_tf_get_fixture):
        axis_title, expected_value = has_tf_get_fixture
        value = axis_title.has_text_frame
        assert value is expected_value

    def it_can_change_whether_it_has_a_text_frame(self, has_tf_set_fixture):
        axis_title, value, expected_xml = has_tf_set_fixture
        axis_title.has_text_frame = value
        assert axis_title._element.xml == expected_xml

    def it_provides_access_to_its_format(self, format_fixture):
        axis_title, ChartFormat_, format_ = format_fixture
        format = axis_title.format
        ChartFormat_.assert_called_once_with(axis_title._element)
        assert format is format_

    def it_provides_access_to_its_text_frame(self, text_frame_fixture):
        axis_title, TextFrame_, text_frame_ = text_frame_fixture
        text_frame = axis_title.text_frame
        TextFrame_.assert_called_once_with(axis_title._element.tx.rich, axis_title)
        assert text_frame is text_frame_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def format_fixture(self, request, ChartFormat_, format_):
        axis_title = AxisTitle(element("c:title"))
        return axis_title, ChartFormat_, format_

    @pytest.fixture(
        params=[
            ("c:title", False),
            ("c:title/c:tx", False),
            ("c:title/c:tx/c:strRef", False),
            ("c:title/c:tx/c:rich", True),
        ]
    )
    def has_tf_get_fixture(self, request):
        title_cxml, expected_value = request.param
        axis_title = AxisTitle(element(title_cxml))
        return axis_title, expected_value

    @pytest.fixture(
        params=[
            (
                "c:title{a:b=c}",
                True,
                "c:title{a:b=c}/c:tx/c:rich/(a:bodyPr,a:lstStyle,a:p/a:pPr/a:defRPr"
                ")",
            ),
            (
                "c:title{a:b=c}/c:tx",
                True,
                "c:title{a:b=c}/c:tx/c:rich/(a:bodyPr,a:lstStyle,a:p/a:pPr/a:defRPr"
                ")",
            ),
            (
                "c:title{a:b=c}/c:tx/c:strRef",
                True,
                "c:title{a:b=c}/c:tx/c:rich/(a:bodyPr,a:lstStyle,a:p/a:pPr/a:defRPr"
                ")",
            ),
            ("c:title/c:tx/c:rich", True, "c:title/c:tx/c:rich"),
            ("c:title", False, "c:title"),
            ("c:title/c:tx", False, "c:title"),
            ("c:title/c:tx/c:rich", False, "c:title"),
            ("c:title/c:tx/c:strRef", False, "c:title"),
        ]
    )
    def has_tf_set_fixture(self, request):
        title_cxml, value, expected_cxml = request.param
        axis_title = AxisTitle(element(title_cxml))
        expected_xml = xml(expected_cxml)
        return axis_title, value, expected_xml

    @pytest.fixture
    def text_frame_fixture(self, request, TextFrame_):
        axis_title = AxisTitle(element("c:title"))
        text_frame_ = TextFrame_.return_value
        return axis_title, TextFrame_, text_frame_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def ChartFormat_(self, request, format_):
        return class_mock(request, "pptx.chart.axis.ChartFormat", return_value=format_)

    @pytest.fixture
    def format_(self, request):
        return instance_mock(request, ChartFormat)

    @pytest.fixture
    def TextFrame_(self, request):
        return class_mock(request, "pptx.chart.axis.TextFrame")


class DescribeCategoryAxis(object):
    def it_knows_its_category_type(self, cat_type_get_fixture):
        category_axis, expected_value = cat_type_get_fixture
        assert category_axis.category_type is expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def cat_type_get_fixture(self):
        category_axis = CategoryAxis(None)
        expected_value = XL_CATEGORY_TYPE.CATEGORY_SCALE
        return category_axis, expected_value


class DescribeDateAxis(object):
    def it_knows_its_category_type(self, cat_type_get_fixture):
        date_axis, expected_value = cat_type_get_fixture
        assert date_axis.category_type is expected_value

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def cat_type_get_fixture(self):
        date_axis = DateAxis(None)
        expected_value = XL_CATEGORY_TYPE.TIME_SCALE
        return date_axis, expected_value


class DescribeMajorGridlines(object):
    def it_provides_access_to_its_format(self, format_fixture):
        gridlines, expected_xml, ChartFormat_, format_ = format_fixture
        format = gridlines.format
        assert gridlines._xAx.xml == expected_xml
        ChartFormat_.assert_called_once_with(
            gridlines._xAx.xpath("c:majorGridlines")[0]
        )
        assert format is format_

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("c:valAx", "c:valAx/c:majorGridlines"),
            ("c:catAx/c:majorGridlines", "c:catAx/c:majorGridlines"),
        ]
    )
    def format_fixture(self, request, ChartFormat_, format_):
        xAx_cxml, expected_cxml = request.param
        gridlines = MajorGridlines(element(xAx_cxml))
        expected_xml = xml(expected_cxml)
        return gridlines, expected_xml, ChartFormat_, format_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def ChartFormat_(self, request, format_):
        return class_mock(request, "pptx.chart.axis.ChartFormat", return_value=format_)

    @pytest.fixture
    def format_(self, request):
        return instance_mock(request, ChartFormat)


class DescribeTickLabels(object):
    def it_provides_access_to_its_font(self, font_fixture):
        tick_labels, Font_, defRPr, font_ = font_fixture
        font = tick_labels.font
        Font_.assert_called_once_with(defRPr)
        assert font is font_

    def it_adds_a_txPr_to_help_font(self, txPr_fixture):
        tick_labels, expected_xml = txPr_fixture
        tick_labels.font
        assert tick_labels._element.xml == expected_xml

    def it_knows_its_number_format(self, number_format_get_fixture):
        tick_labels, expected_value = number_format_get_fixture
        assert tick_labels.number_format == expected_value

    def it_can_change_its_number_format(self, number_format_set_fixture):
        tick_labels, new_value, expected_xml = number_format_set_fixture
        tick_labels.number_format = new_value
        assert tick_labels._element.xml == expected_xml

    def it_knows_whether_its_number_format_is_linked(
        self, number_format_is_linked_get_fixture
    ):
        tick_labels, expected_value = number_format_is_linked_get_fixture
        assert tick_labels.number_format_is_linked is expected_value

    def it_can_change_whether_its_number_format_is_linked(
        self, number_format_is_linked_set_fixture
    ):
        tick_labels, new_value, expected_xml = number_format_is_linked_set_fixture
        tick_labels.number_format_is_linked = new_value
        assert tick_labels._element.xml == expected_xml

    def it_knows_its_offset(self, offset_get_fixture):
        tick_labels, expected_value = offset_get_fixture
        assert tick_labels.offset == expected_value

    def it_can_change_its_offset(self, offset_set_fixture):
        tick_labels, new_value, expected_xml = offset_set_fixture
        tick_labels.offset = new_value
        assert tick_labels._element.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def font_fixture(self, Font_, font_):
        catAx = element("c:catAx/c:txPr/a:p/a:pPr/a:defRPr")
        defRPr = catAx.xpath(".//a:defRPr")[0]
        tick_labels = TickLabels(catAx)
        return tick_labels, Font_, defRPr, font_

    @pytest.fixture(
        params=[
            ("c:catAx", "General"),
            ("c:valAx/c:numFmt{formatCode=General}", "General"),
        ]
    )
    def number_format_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        tick_labels = TickLabels(element(xAx_cxml))
        return tick_labels, expected_value

    @pytest.fixture(
        params=[
            (
                "c:catAx",
                "General",
                "c:catAx/c:numFmt{formatCode=General,sourceLinked=0}",
            ),
            (
                "c:valAx/c:numFmt{formatCode=General}",
                "00.00",
                "c:valAx/c:numFmt{formatCode=00.00,sourceLinked=0}",
            ),
        ]
    )
    def number_format_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        tick_labels = TickLabels(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return tick_labels, new_value, expected_xml

    @pytest.fixture(
        params=[
            ("c:catAx", False),
            ("c:valAx/c:numFmt", True),
            ("c:valAx/c:numFmt{sourceLinked=0}", False),
            ("c:catAx/c:numFmt{sourceLinked=1}", True),
        ]
    )
    def number_format_is_linked_get_fixture(self, request):
        xAx_cxml, expected_value = request.param
        tick_labels = TickLabels(element(xAx_cxml))
        return tick_labels, expected_value

    @pytest.fixture(
        params=[
            ("c:valAx", True, "c:valAx/c:numFmt{sourceLinked=1}"),
            ("c:catAx", False, "c:catAx/c:numFmt{sourceLinked=0}"),
            ("c:valAx", None, "c:valAx/c:numFmt"),
            ("c:catAx/c:numFmt", True, "c:catAx/c:numFmt{sourceLinked=1}"),
            (
                "c:valAx/c:numFmt{sourceLinked=1}",
                False,
                "c:valAx/c:numFmt{sourceLinked=0}",
            ),
        ]
    )
    def number_format_is_linked_set_fixture(self, request):
        xAx_cxml, new_value, expected_xAx_cxml = request.param
        tick_labels = TickLabels(element(xAx_cxml))
        expected_xml = xml(expected_xAx_cxml)
        return tick_labels, new_value, expected_xml

    @pytest.fixture(
        params=[
            ("c:catAx", 100),
            ("c:catAx/c:lblOffset", 100),
            ("c:catAx/c:lblOffset{val=420}", 420),
            ("c:catAx/c:lblOffset{val=004}", 4),
            ("c:catAx/c:lblOffset{val=42%}", 42),
            ("c:catAx/c:lblOffset{val=02%}", 2),
        ]
    )
    def offset_get_fixture(self, request):
        catAx_cxml, expected_value = request.param
        tick_labels = TickLabels(element(catAx_cxml))
        return tick_labels, expected_value

    @pytest.fixture(
        params=[
            ("c:catAx", 420, "c:catAx/c:lblOffset{val=420}"),
            ("c:catAx/c:lblOffset{val=420}", 100, "c:catAx"),
        ]
    )
    def offset_set_fixture(self, request):
        catAx_cxml, new_value, expected_catAx_cxml = request.param
        tick_labels = TickLabels(element(catAx_cxml))
        expected_xml = xml(expected_catAx_cxml)
        return tick_labels, new_value, expected_xml

    @pytest.fixture(
        params=[
            (
                "c:valAx{a:b=c}",
                "c:valAx{a:b=c}/c:txPr/(a:bodyPr,a:lstStyle,a:p/a:pPr/a:defRPr)",
            ),
            (
                "c:valAx{a:b=c}/c:txPr/(a:bodyPr,a:p)",
                "c:valAx{a:b=c}/c:txPr/(a:bodyPr,a:p/a:pPr/a:defRPr)",
            ),
            (
                "c:valAx{a:b=c}/c:txPr/(a:bodyPr,a:p/a:pPr)",
                "c:valAx{a:b=c}/c:txPr/(a:bodyPr,a:p/a:pPr/a:defRPr)",
            ),
        ]
    )
    def txPr_fixture(self, request):
        xAx_cxml, expected_cxml = request.param
        tick_labels = TickLabels(element(xAx_cxml))
        expected_xml = xml(expected_cxml)
        return tick_labels, expected_xml

    # fixture components ---------------------------------------------

    @pytest.fixture
    def Font_(self, request, font_):
        return class_mock(request, "pptx.chart.axis.Font", return_value=font_)

    @pytest.fixture
    def font_(self, request):
        return instance_mock(request, Font)


class DescribeValueAxis(object):
    def it_knows_the_other_axis_crossing_type(self, crosses_get_fixture):
        value_axis, expected_value = crosses_get_fixture
        assert value_axis.crosses == expected_value

    def it_can_change_the_other_axis_crossing_type(self, crosses_set_fixture):
        value_axis, new_value, plotArea, expected_xml = crosses_set_fixture
        value_axis.crosses = new_value
        assert plotArea.xml == expected_xml

    def it_knows_the_other_axis_crossing_value(self, crosses_at_get_fixture):
        value_axis, expected_value = crosses_at_get_fixture
        assert value_axis.crosses_at == expected_value

    def it_can_change_the_other_axis_crossing_value(self, crosses_at_set_fixture):
        value_axis, new_value, plotArea, expected_xml = crosses_at_set_fixture
        value_axis.crosses_at = new_value
        assert plotArea.xml == expected_xml

    def it_knows_its_major_unit(self, major_unit_get_fixture):
        value_axis, expected_value = major_unit_get_fixture
        assert value_axis.major_unit == expected_value

    def it_can_change_its_major_unit(self, major_unit_set_fixture):
        value_axis, new_value, expected_xml = major_unit_set_fixture
        value_axis.major_unit = new_value
        assert value_axis._element.xml == expected_xml

    def it_knows_its_minor_unit(self, minor_unit_get_fixture):
        value_axis, expected_value = minor_unit_get_fixture
        assert value_axis.minor_unit == expected_value

    def it_can_change_its_minor_unit(self, minor_unit_set_fixture):
        value_axis, new_value, expected_xml = minor_unit_set_fixture
        value_axis.minor_unit = new_value
        assert value_axis._element.xml == expected_xml

    # fixtures -------------------------------------------------------

    @pytest.fixture(
        params=[
            ("c:plotArea/(c:valAx/c:axId{val=42},c:valAx/c:crossAx{val=42})", None),
            (
                "c:plotArea/(c:catAx/(c:axId{val=42},c:crossesAt{val=2.4}),c:valAx/"
                "c:crossAx{val=42})",
                2.4,
            ),
            (
                "c:plotArea/(c:dateAx/(c:axId{val=42},c:crossesAt{val=-1.2}),c:valA"
                "x/c:crossAx{val=42})",
                -1.2,
            ),
        ]
    )
    def crosses_at_get_fixture(self, request):
        plotArea_cxml, expected_value = request.param
        plotArea = element(plotArea_cxml)
        valAx = plotArea.xpath('c:valAx[c:crossAx/@val="42"]')[0]
        value_axis = ValueAxis(valAx)
        return value_axis, expected_value

    @pytest.fixture(
        params=[
            (
                "c:plotArea/(c:valAx/c:axId{val=42},c:valAx/c:crossAx{val=42})",
                2.4,
                "c:plotArea/(c:valAx/(c:axId{val=42},c:crossesAt{val=2.4}),c:valAx/"
                "c:crossAx{val=42})",
            ),
            (
                "c:plotArea/(c:catAx/(c:axId{val=42},c:crosses{val=min}),c:valAx/c:"
                "crossAx{val=42})",
                1.5,
                "c:plotArea/(c:catAx/(c:axId{val=42},c:crossesAt{val=1.5}),c:valAx/"
                "c:crossAx{val=42})",
            ),
            (
                "c:plotArea/(c:dateAx/(c:axId{val=42},c:crossesAt{val=2.4}),c:valAx"
                "/c:crossAx{val=42})",
                1.5,
                "c:plotArea/(c:dateAx/(c:axId{val=42},c:crossesAt{val=1.5}),c:valAx"
                "/c:crossAx{val=42})",
            ),
            (
                "c:plotArea/(c:catAx/(c:axId{val=42},c:crossesAt{val=1.5}),c:valAx/"
                "c:crossAx{val=42})",
                None,
                "c:plotArea/(c:catAx/(c:axId{val=42}),c:valAx/c:crossAx{val=42})",
            ),
        ]
    )
    def crosses_at_set_fixture(self, request):
        plotArea_cxml, new_value, expected_cxml = request.param
        plotArea = element(plotArea_cxml)
        valAx = plotArea.xpath('c:valAx[c:crossAx/@val="42"]')[0]
        value_axis = ValueAxis(valAx)
        expected_xml = xml(expected_cxml)
        return value_axis, new_value, plotArea, expected_xml

    @pytest.fixture(
        params=[
            ("c:plotArea/(c:valAx/c:axId{val=42},c:valAx/c:crossAx{val=42})", "CUSTOM"),
            (
                "c:plotArea/(c:catAx/(c:axId{val=42},c:crosses{val=autoZero}),c:val"
                "Ax/c:crossAx{val=42})",
                "AUTOMATIC",
            ),
            (
                "c:plotArea/(c:valAx/(c:axId{val=42},c:crosses{val=min}),c:valAx/c:"
                "crossAx{val=42})",
                "MINIMUM",
            ),
        ]
    )
    def crosses_get_fixture(self, request):
        cxml, member = request.param
        valAx = element(cxml).xpath('c:valAx[c:crossAx/@val="42"]')[0]
        value_axis = ValueAxis(valAx)
        expected_value = getattr(XL_AXIS_CROSSES, member)
        return value_axis, expected_value

    @pytest.fixture(
        params=[
            (
                "c:plotArea/(c:valAx/(c:axId{val=42},c:crossesAt{val=2.4}),c:valAx/"
                "c:crossAx{val=42})",
                "AUTOMATIC",
                "c:plotArea/(c:valAx/(c:axId{val=42},c:crosses{val=autoZero}),c:val"
                "Ax/c:crossAx{val=42})",
            ),
            (
                "c:plotArea/(c:catAx/(c:axId{val=42},c:crosses{val=autoZero}),c:val"
                "Ax/c:crossAx{val=42})",
                "MINIMUM",
                "c:plotArea/(c:catAx/(c:axId{val=42},c:crosses{val=min}),c:valAx/c:"
                "crossAx{val=42})",
            ),
            (
                "c:plotArea/(c:valAx/(c:axId{val=42},c:crosses{val=min}),c:valAx/c:"
                "crossAx{val=42})",
                "CUSTOM",
                "c:plotArea/(c:valAx/(c:axId{val=42},c:crossesAt{val=0.0}),c:valAx/"
                "c:crossAx{val=42})",
            ),
            (
                "c:plotArea/(c:catAx/(c:axId{val=42},c:crossesAt{val=2.4}),c:valAx/"
                "c:crossAx{val=42})",
                "CUSTOM",
                "c:plotArea/(c:catAx/(c:axId{val=42},c:crossesAt{val=2.4}),c:valAx/"
                "c:crossAx{val=42})",
            ),
        ]
    )
    def crosses_set_fixture(self, request):
        plotArea_cxml, member, expected_cxml = request.param
        plotArea = element(plotArea_cxml)
        valAx = plotArea.xpath('c:valAx[c:crossAx/@val="42"]')[0]
        value_axis = ValueAxis(valAx)
        new_value = getattr(XL_AXIS_CROSSES, member)
        expected_xml = xml(expected_cxml)
        return value_axis, new_value, plotArea, expected_xml

    @pytest.fixture(params=[("c:valAx", None), ("c:valAx/c:majorUnit{val=4.2}", 4.2)])
    def major_unit_get_fixture(self, request):
        valAx_cxml, expected_value = request.param
        value_axis = ValueAxis(element(valAx_cxml))
        return value_axis, expected_value

    @pytest.fixture(
        params=[
            ("c:valAx", 42, "c:valAx/c:majorUnit{val=42.0}"),
            ("c:valAx", None, "c:valAx"),
            ("c:valAx/c:majorUnit{val=42.0}", 24.0, "c:valAx/c:majorUnit{val=24.0}"),
            ("c:valAx/c:majorUnit{val=42.0}", None, "c:valAx"),
        ]
    )
    def major_unit_set_fixture(self, request):
        valAx_cxml, new_value, expected_valAx_cxml = request.param
        value_axis = ValueAxis(element(valAx_cxml))
        expected_xml = xml(expected_valAx_cxml)
        return value_axis, new_value, expected_xml

    @pytest.fixture(params=[("c:valAx", None), ("c:valAx/c:minorUnit{val=2.4}", 2.4)])
    def minor_unit_get_fixture(self, request):
        valAx_cxml, expected_value = request.param
        value_axis = ValueAxis(element(valAx_cxml))
        return value_axis, expected_value

    @pytest.fixture(
        params=[
            ("c:valAx", 36, "c:valAx/c:minorUnit{val=36.0}"),
            ("c:valAx", None, "c:valAx"),
            ("c:valAx/c:minorUnit{val=36.0}", 12.6, "c:valAx/c:minorUnit{val=12.6}"),
            ("c:valAx/c:minorUnit{val=36.0}", None, "c:valAx"),
        ]
    )
    def minor_unit_set_fixture(self, request):
        valAx_cxml, new_value, expected_valAx_cxml = request.param
        value_axis = ValueAxis(element(valAx_cxml))
        expected_xml = xml(expected_valAx_cxml)
        return value_axis, new_value, expected_xml



# encoding: utf-8

"""
Test suite for pptx.api module
"""

from __future__ import absolute_import, division, print_function, unicode_literals

import os

import pytest

from pptx.api import Presentation
from pptx.opc.constants import CONTENT_TYPE as CT
from pptx.parts.presentation import PresentationPart

from .unitutil.mock import class_mock, instance_mock


class DescribePresentation(object):
    def it_opens_default_template_on_no_path_provided(self, call_fixture):
        Package_, path, prs_ = call_fixture
        prs = Presentation()
        Package_.open.assert_called_once_with(path)
        assert prs is prs_

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def call_fixture(self, Package_, prs_, prs_part_):
        path = os.path.abspath(
            os.path.join(
                os.path.split(__file__)[0], "../pptx/templates", "default.pptx"
            )
        )
        Package_.open.return_value.main_document_part = prs_part_
        prs_part_.content_type = CT.PML_PRESENTATION_MAIN
        prs_part_.presentation = prs_
        return Package_, path, prs_

    # fixture components ---------------------------------------------

    @pytest.fixture
    def Package_(self, request):
        return class_mock(request, "pptx.api.Package")

    @pytest.fixture
    def prs_(self, request):
        return instance_mock(request, Presentation)

    @pytest.fixture
    def prs_part_(self, request):
        return instance_mock(request, PresentationPart)



# encoding: utf-8

"""Gherkin step implementations for chart plot features."""

from __future__ import absolute_import, division, print_function, unicode_literals

from behave import given, then, when

from pptx import Presentation

from helpers import test_pptx


# given ===================================================


@given("a bar plot {having_or_not} data labels")
def given_a_bar_plot_having_or_not_data_labels(context, having_or_not):
    slide_idx = {"having": 0, "not having": 1}[having_or_not]
    prs = Presentation(test_pptx("cht-plot-props"))
    context.plot = prs.slides[slide_idx].shapes[0].chart.plots[0]


@given("a bar plot having gap width of {width}")
def given_a_bar_plot_having_gap_width_of_width(context, width):
    slide_idx = {"no explicit value": 0, "300": 1}[width]
    prs = Presentation(test_pptx("cht-plot-props"))
    context.plot = prs.slides[slide_idx].shapes[0].chart.plots[0]


@given("a bar plot having overlap of {overlap}")
def given_a_bar_plot_having_overlap_of_overlap(context, overlap):
    slide_idx = {"no explicit value": 0, "42": 1, "-42": 2}[overlap]
    prs = Presentation(test_pptx("cht-plot-props"))
    context.plot = prs.slides[slide_idx].shapes[0].chart.plots[0]


@given("a bar plot having vary color by category set to {setting}")
def given_a_bar_plot_having_vary_color_by_category_setting(context, setting):
    slide_idx = {"no explicit setting": 0, "True": 1, "False": 2}[setting]
    prs = Presentation(test_pptx("cht-plot-props"))
    context.plot = prs.slides[slide_idx].shapes[0].chart.plots[0]


@given("a bubble plot having bubble scale of {percent}")
def given_a_bubble_plot_having_bubble_scale_of_percent(context, percent):
    slide_idx = {"no explicit value": 3, "70%": 4}[percent]
    prs = Presentation(test_pptx("cht-plot-props"))
    context.bubble_plot = prs.slides[slide_idx].shapes[0].chart.plots[0]


@given("a category plot")
def given_a_category_plot(context):
    prs = Presentation(test_pptx("cht-plot-props"))
    context.plot = prs.slides[2].shapes[0].chart.plots[0]


# when ====================================================


@when("I assign {value} to bubble_plot.bubble_scale")
def when_I_assign_value_to_bubble_plot_bubble_scale(context, value):
    new_value = None if value == "None" else int(value)
    context.bubble_plot.bubble_scale = new_value


@when("I assign {value} to plot.gap_width")
def when_I_assign_value_to_plot_gap_width(context, value):
    new_value = int(value)
    context.plot.gap_width = new_value


@when("I assign {value} to plot.has_data_labels")
def when_I_assign_value_to_plot_has_data_labels(context, value):
    new_value = {"True": True, "False": False}[value]
    context.plot.has_data_labels = new_value


@when("I assign {value} to plot.overlap")
def when_I_assign_value_to_plot_overlap(context, value):
    new_value = int(value)
    context.plot.overlap = new_value


@when("I assign {value} to plot.vary_by_categories")
def when_I_assign_value_to_plot_vary_by_categories(context, value):
    new_value = {"True": True, "False": False}[value]
    context.plot.vary_by_categories = new_value


# then ====================================================


@then("bubble_plot.bubble_scale is {value}")
def then_bubble_plot_bubble_scale_is_value(context, value):
    expected_value = int(value)
    bubble_plot = context.bubble_plot
    assert bubble_plot.bubble_scale == expected_value, (
        "got %s" % bubble_plot.bubble_scale
    )


@then("len(plot.categories) is {count}")
def then_len_plot_categories_is_count(context, count):
    plot = context.chart.plots[0]
    expected_count = int(count)
    assert len(plot.categories) == expected_count


@then("plot.categories is a Categories object")
def then_plot_categories_is_a_Categories_object(context):
    plot = context.plot
    type_name = type(plot.categories).__name__
    assert type_name == "Categories", "got %s" % type_name


@then("plot.gap_width is {value}")
def then_plot_gap_width_is_value(context, value):
    expected_value = int(value)
    plot = context.plot
    assert plot.gap_width == expected_value, "got %s" % plot.gap_width


@then("plot.has_data_labels is {value}")
def then_plot_has_data_labels_is_value(context, value):
    expected_value = {"True": True, "False": False}[value]
    assert context.plot.has_data_labels is expected_value


@then("plot.overlap is {value}")
def then_plot_overlap_is_expected_value(context, value):
    expected_value = int(value)
    plot = context.plot
    assert plot.overlap == expected_value, "got %s" % plot.overlap


@then("plot.vary_by_categories is {value}")
def then_plot_vary_by_categories_is_value(context, value):
    expected_value = {"True": True, "False": False}[value]
    plot = context.plot
    assert plot.vary_by_categories is expected_value



# encoding: utf-8

"""
API classes for dealing with presentations and other objects one typically
encounters as an end-user of the PowerPoint user interface.
"""

from __future__ import absolute_import, division, print_function, unicode_literals

from .opc.constants import RELATIONSHIP_TYPE as RT
from .opc.package import OpcPackage
from .opc.packuri import PackURI
from .parts.coreprops import CorePropertiesPart
from .parts.image import Image, ImagePart
from .parts.media import MediaPart
from .util import lazyproperty


class Package(OpcPackage):
    """
    Return an instance of |Package| loaded from *file*, where *file* can be a
    path (a string) or a file-like object. If *file* is a path, it can be
    either a path to a PowerPoint `.pptx` file or a path to a directory
    containing an expanded presentation file, as would result from unzipping
    a `.pptx` file. If *file* is |None|, the default presentation template is
    loaded.
    """

    @lazyproperty
    def core_properties(self):
        """
        Instance of |CoreProperties| holding the read/write Dublin Core
        document properties for this presentation. Creates a default core
        properties part if one is not present (not common).
        """
        try:
            return self.part_related_by(RT.CORE_PROPERTIES)
        except KeyError:
            core_props = CorePropertiesPart.default()
            self.relate_to(core_props, RT.CORE_PROPERTIES)
            return core_props

    def get_or_add_image_part(self, image_file):
        """
        Return an |ImagePart| object containing the image in *image_file*. If
        the image part already exists in this package, it is reused,
        otherwise a new one is created.
        """
        return self._image_parts.get_or_add_image_part(image_file)

    def get_or_add_media_part(self, media):
        """Return a |MediaPart| object containing the media in *media*.

        If a media part for this media bytestream ("file") is already present
        in this package, it is reused, otherwise a new one is created.
        """
        return self._media_parts.get_or_add_media_part(media)

    def next_image_partname(self, ext):
        """
        Return a |PackURI| instance representing the next available image
        partname, by sequence number. *ext* is used as the extention on the
        returned partname.
        """

        def first_available_image_idx():
            image_idxs = sorted(
                [
                    part.partname.idx
                    for part in self.iter_parts()
                    if part.partname.startswith("/ppt/media/image")
                    and part.partname.idx is not None
                ]
            )
            for i, image_idx in enumerate(image_idxs):
                idx = i + 1
                if idx < image_idx:
                    return idx
            return len(image_idxs) + 1

        idx = first_available_image_idx()
        return PackURI("/ppt/media/image%d.%s" % (idx, ext))

    def next_media_partname(self, ext):
        """Return |PackURI| instance for next available media partname.

        Partname is first available, starting at sequence number 1. Empty
        sequence numbers are reused. *ext* is used as the extension on the
        returned partname.
        """

        def first_available_media_idx():
            media_idxs = sorted(
                [
                    part.partname.idx
                    for part in self.iter_parts()
                    if part.partname.startswith("/ppt/media/media")
                ]
            )
            for i, media_idx in enumerate(media_idxs):
                idx = i + 1
                if idx < media_idx:
                    return idx
            return len(media_idxs) + 1

        idx = first_available_media_idx()
        return PackURI("/ppt/media/media%d.%s" % (idx, ext))

    @property
    def presentation_part(self):
        """
        Reference to the |Presentation| instance contained in this package.
        """
        return self.main_document_part

    @lazyproperty
    def _image_parts(self):
        """
        |_ImageParts| object providing access to the image parts in this
        package.
        """
        return _ImageParts(self)

    @lazyproperty
    def _media_parts(self):
        """Return |_MediaParts| object for this package.

        The media parts object provides access to all the media parts in this
        package.
        """
        return _MediaParts(self)


class _ImageParts(object):
    """Provides access to the image parts in a package."""

    def __init__(self, package):
        super(_ImageParts, self).__init__()
        self._package = package

    def __iter__(self):
        """
        Generate a reference to each |ImagePart| object in the package.
        """
        image_parts = []
        for rel in self._package.iter_rels():
            if rel.is_external:
                continue
            if rel.reltype != RT.IMAGE:
                continue
            image_part = rel.target_part
            if image_part in image_parts:
                continue
            image_parts.append(image_part)
            yield image_part

    def get_or_add_image_part(self, image_file):
        """
        Return an |ImagePart| object containing the image in *image_file*,
        which is either a path to an image file or a file-like object
        containing an image. If an image part containing this same image
        already exists, that instance is returned, otherwise a new image part
        is created.
        """
        image = Image.from_file(image_file)
        image_part = self._find_by_sha1(image.sha1)
        if image_part is None:
            image_part = ImagePart.new(self._package, image)
        return image_part

    def _find_by_sha1(self, sha1):
        """
        Return an |ImagePart| object belonging to this package or |None| if
        no matching image part is found. The image part is identified by the
        SHA1 hash digest of the image binary it contains.
        """
        for image_part in self:
            # ---skip unknown/unsupported image types, like SVG---
            if not hasattr(image_part, "sha1"):
                continue
            if image_part.sha1 == sha1:
                return image_part
        return None


class _MediaParts(object):
    """Provides access to the media parts in a package.

    Supports iteration and :meth:`get()` using the media object SHA1 hash as
    its key.
    """

    def __init__(self, package):
        super(_MediaParts, self).__init__()
        self._package = package

    def __iter__(self):
        """Generate a reference to each |MediaPart| object in the package."""
        # A media part can appear in more than one relationship (and commonly
        # does in the case of video). Use media_parts to keep track of those
        # that have been "yielded"; they can be skipped if they occur again.
        media_parts = []
        for rel in self._package.iter_rels():
            if rel.is_external:
                continue
            if rel.reltype not in (RT.MEDIA, RT.VIDEO):
                continue
            media_part = rel.target_part
            if media_part in media_parts:
                continue
            media_parts.append(media_part)
            yield media_part

    def get_or_add_media_part(self, media):
        """Return a |MediaPart| object containing the media in *media*.

        If this package already contains a media part for the same
        bytestream, that instance is returned, otherwise a new media part is
        created.
        """
        media_part = self._find_by_sha1(media.sha1)
        if media_part is None:
            media_part = MediaPart.new(self._package, media)
        return media_part

    def _find_by_sha1(self, sha1):
        """Return |MediaPart| object having *sha1* hash or None if not found.

        All media parts belonging to this package are considered. A media
        part is identified by the SHA1 hash digest of its bytestream
        ("file").
        """
        for media_part in self:
            if media_part.sha1 == sha1:
                return media_part
        return None



# encoding: utf-8

"""Gherkin step implementations for table-related features"""

from __future__ import absolute_import, division, print_function, unicode_literals

from behave import given, when, then

from pptx import Presentation
from pptx.enum.text import MSO_ANCHOR  # noqa
from pptx.util import Inches

from helpers import test_pptx


# given ===================================================


@given("a Table object as table")
@given("a 2x2 Table object as table")
def given_a_2x2_Table_object_as_table(context):
    prs = Presentation(test_pptx("shp-shapes"))
    context.table_ = prs.slides[0].shapes[3].table


@given("a 2x3 _MergeOriginCell object as cell")
def given_a_2x3_MergeOriginCell_object_as_cell(context):
    prs = Presentation(test_pptx("tbl-cell"))
    context.cell = prs.slides[1].shapes[1].table.cell(0, 0)


@given("a 3x3 Table object as table")
@given("a 3x3 Table object with cells a to i as table")
def given_a_3x3_table_with_cells_a_to_i_as_table(context):
    prs = Presentation(test_pptx("tbl-cell"))
    # ---context.table is used by Behave for some odd reason---
    context.table_ = prs.slides[2].shapes[0].table


@given("a _Cell object as cell")
def given_a_Cell_object_as_cell(context):
    prs = Presentation(test_pptx("shp-shapes"))
    context.cell = prs.slides[0].shapes[3].table.cell(0, 0)


@given('a _Cell object containing "unladen swallows" as cell')
def given_a_Cell_object_containing_unladen_swallows_as_cell(context):
    prs = Presentation(test_pptx("tbl-cell"))
    context.cell = prs.slides[0].shapes[0].table.cell(1, 0)


@given("a _Cell object with known margins as cell")
def given_a_Cell_object_with_known_margins_as_cell(context):
    prs = Presentation(test_pptx("tbl-cell"))
    context.cell = prs.slides[0].shapes[0].table.cell(0, 0)


@given("a _Cell object with {setting} vertical alignment as cell")
def given_a_Cell_object_with_setting_vertical_alignment(context, setting):
    cell_coordinates = {"inherited": (0, 1), "middle": (0, 2), "bottom": (0, 3)}[
        setting
    ]
    prs = Presentation(test_pptx("tbl-cell"))
    context.cell = prs.slides[0].shapes[0].table.cell(*cell_coordinates)


@given("a {role} _Cell object as cell")
def given_a_role_Cell_object_as_cell(context, role):
    coordinates = {"merge-origin": (0, 0), "spanned": (0, 1), "unmerged": (2, 2)}[role]
    table = Presentation(test_pptx("tbl-cell")).slides[1].shapes[0].table
    # ---create other_cell here where we know the coordinates---
    context.cell = table.cell(*coordinates)
    context.other_cell = table.cell(*coordinates)


@given("a second proxy instance for that cell as other_cell")
def given_a_second_proxy_instance_for_that_cell_as_other_cell(context):
    # ---other_cell is actually produced by prior step---
    assert context.other_cell


@given("a _Column object as column")
def given_a_Column_object_as_column(context):
    prs = Presentation(test_pptx("shp-shapes"))
    context.column = prs.slides[0].shapes[3].table.columns[0]


# when ====================================================


@when("I assign cell.margin_{side} = {value}")
def when_I_assign_cell_margin_side_eq_value(context, value, side):
    setattr(context.cell, "margin_%s" % side, eval(value))


@when('I assign cell.text = "test text"')
def when_I_assign_cell_text(context):
    context.cell.text = "test text"


@when("I assign cell.vertical_anchor = {value}")
def when_I_assign_cell_vertical_anchor_eq_value(context, value):
    context.cell.vertical_anchor = eval(value)


@when("I assign column.width = {value}")
def when_I_assign_column_width_eq_value(context, value):
    context.column.width = eval(value)


@when("I assign origin_cell = table.cell(0, 0)")
def when_I_assign_origin_cell_eq_table_cell_0_0(context):
    context.origin_cell = context.table_.cell(0, 0)


@when("I assign other_cell = table.cell(1, 1)")
def when_I_assign_other_cell_eq_table_cell_1_1(context):
    context.other_cell = context.table_.cell(1, 1)


@when("I assign table.first_col = True")
def when_I_assign_table_first_col_eq_True(context):
    context.table_.first_col = True


@when("I assign table.first_row = True")
def when_I_assign_table_first_row_eq_True(context):
    context.table_.first_row = True


@when("I assign table.horz_banding = True")
def when_I_assign_table_horz_banding_eq_True(context):
    context.table_.horz_banding = True


@when("I assign table.last_col = True")
def when_I_assign_table_last_col_eq_True(context):
    context.table_.last_col = True


@when("I assign table.last_row = True")
def when_I_assign_table_last_row_eq_True(context):
    context.table_.last_row = True


@when("I assign table.vert_banding = True")
def when_I_assign_table_vert_banding_eq_True(context):
    context.table_.vert_banding = True


@when("I call cell.split()")
def when_I_call_cell_split_other_cell(context):
    context.cell.split()


@when("I call origin_cell.merge(other_cell)")
def when_I_call_origin_cell_merge_other_cell(context):
    context.origin_cell.merge(context.other_cell)


# then ====================================================


@then("cell == other_cell")
def then_cell_eq_other_cell(context):
    cell, other_cell = context.cell, context.other_cell
    assert cell == other_cell, "cell != other_cell"


@then("cell.fill is a FillFormat object")
def then_cell_fill_is_a_FillFormat_object(context):
    actual = type(context.cell.fill).__name__
    expected = "FillFormat"
    assert actual == expected, "cell.fill is a %s object" % actual


@then("cell.margin_{side} == Inches({num_lit})")
def then_cell_margin_side_eq_Inches_num(context, side, num_lit):
    actual = getattr(context.cell, "margin_%s" % side)
    expected = Inches(float(num_lit))
    assert actual == expected, "cell.margin_%s == %s" % (side, actual.inches)


@then("{cell_ref}.is_merge_origin is {bool_lit}")
def then_cell_ref_is_merge_origin_is(context, cell_ref, bool_lit):
    expected = eval(bool_lit)
    actual = getattr(context, cell_ref).is_merge_origin
    assert actual is expected, "%s.is_merge_origin is %s" % (cell_ref, actual)


@then("{cell_ref}.is_spanned is {bool_lit}")
def then_cell_is_spanned_is(context, cell_ref, bool_lit):
    expected = eval(bool_lit)
    actual = getattr(context, cell_ref).is_spanned
    assert actual is expected, "%s.is_spanned is %s" % (cell_ref, actual)


@then("cell.text == {value}")
def then_cell_text_eq_value(context, value):
    actual, expected = context.cell.text, eval(value)
    assert actual == expected, 'cell.text == "%s"' % actual


@then("cell.span_height == {int_lit}")
def then_cell_span_height_eq(context, int_lit):
    expected = int(int_lit)
    actual = context.cell.span_height
    assert actual is expected, "cell.span_height == %s" % actual


@then("cell.span_width == {int_lit}")
def then_cell_span_width_eq(context, int_lit):
    expected = int(int_lit)
    actual = context.cell.span_width
    assert actual is expected, "cell.span_width == %s" % actual


@then("cell.vertical_anchor == {value}")
def then_cell_vertical_anchor_eq_value(context, value):
    actual = context.cell.vertical_anchor
    expected = eval(value)
    assert actual == expected, "cell.vertical_anchor == %s" % actual


@then("column.width.inches == {float_lit}")
def then_column_width_inches_eq(context, float_lit):
    actual = context.column.width.inches
    expected = float(float_lit)
    assert actual == expected, "column.width.inches == %s" % actual


@then("len(list(table.iter_cells())) == {int_lit}")
def then_len_list_table_iter_cells_eq(context, int_lit):
    actual = len(list(context.table_.iter_cells()))
    expected = int(int_lit)
    assert actual is expected, "len(list(table.iter_cells())) == %s" % actual


@then("origin_cell.text == {value}")
def then_origin_cell_text_eq_value(context, value):
    actual, expected = context.origin_cell.text, eval(value)
    assert actual == expected, 'origin_cell.text == "%s"' % actual


@then("other_cell.text == {value}")
def then_other_cell_text_eq_value(context, value):
    actual, expected = context.other_cell.text, eval(value)
    assert actual == expected, 'other_cell.text == "%s"' % actual


@then("table.cell(0, 0) is a {type_name} object")
def then_table_cell_0_0_is_a_type_object(context, type_name):
    actual = type(context.table_.cell(0, 0)).__name__
    expected = type_name
    assert actual == expected, "table.cell(0, 0) is a %s object" % actual


@then("table.columns is a {type_name} object")
def then_table_columns_is_a_type_object(context, type_name):
    actual = type(context.table_.columns).__name__
    expected = type_name
    assert actual == expected, "table.columns is a %s object" % actual


@then("table.first_col is {bool_lit}")
def then_table_first_col_is_value(context, bool_lit):
    actual = context.table_.first_col
    expected = eval(bool_lit)
    assert actual is expected, "table.first_col is %s" % actual


@then("table.first_row is {bool_lit}")
def then_table_first_row_is_value(context, bool_lit):
    actual = context.table_.first_row
    expected = eval(bool_lit)
    assert actual is expected, "table.first_row is %s" % actual


@then("table.horz_banding is {bool_lit}")
def then_table_horz_banding_is_value(context, bool_lit):
    actual = context.table_.horz_banding
    expected = eval(bool_lit)
    assert actual is expected, "table.horz_banding is %s" % actual


@then("table.last_col is {bool_lit}")
def then_table_last_col_is_value(context, bool_lit):
    actual = context.table_.last_col
    expected = eval(bool_lit)
    assert actual is expected, "table.last_col is %s" % actual


@then("table.last_row is {bool_lit}")
def then_table_last_row_is_value(context, bool_lit):
    actual = context.table_.last_row
    expected = eval(bool_lit)
    assert actual is expected, "table.last_row is %s" % actual


@then("table.rows is a {type_name} object")
def then_table_rows_is_a_type_object(context, type_name):
    actual = type(context.table_.rows).__name__
    expected = type_name
    assert actual == expected, "table.rows is a %s object" % actual


@then("table.vert_banding is {bool_lit}")
def then_table_vert_banding_is_value(context, bool_lit):
    actual = context.table_.vert_banding
    expected = eval(bool_lit)
    assert actual is expected, "table.vert_banding is %s" % actual
